/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: a_grid.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "a_grid.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef WIN32
#include <io.h>
//#define open(a,b,c) _open((a),(b),(c))
//#define close(fd) _close((fd))
#define ftruncate(fd,p) _chsize((fd),(p))
#define strcasecmp stricmp
#define snprintf _snprintf
#else
#include <unistd.h>
#endif

a_grid_t* ag_creategrid() {
  a_grid_t* result  = (a_grid_t*)malloc(sizeof(a_grid_t));
  result->cells     = NULL;
  result->rowcount  = 0;
  result->colcount  = 0;
  return result;
}

void ag_deletegrid(a_grid_t* grid) {
  if (grid) {
    int i, j;
    for (i=0; i<grid->rowcount; i++) {
      for (j=0; j<grid->colcount; j++) {
        if (grid->cells[i][j]) {
          free(grid->cells[i][j]);
        }
      }
      free(grid->cells[i]);
    }
    free(grid->cells);
    free(grid);
  }
}

int ag_addrow(a_grid_t* grid) {
  if (grid) {
    int i;
    grid->cells = (char***)realloc(grid->cells,
                  sizeof(char**) * (grid->rowcount+1));
    grid->cells[grid->rowcount] = (char**)
                                  malloc(sizeof(char*) * (grid->colcount));
    for (i=0; i<grid->colcount; i++) {
      grid->cells[grid->rowcount][i] = NULL;
    }
    return grid->rowcount++;
  }
  return -1;
}

int ag_addcol(a_grid_t* grid) {
  if (grid) {
    int i;
    for (i=0; i<grid->rowcount; i++) {
      grid->cells[i] = (char**)realloc(grid->cells[i],
                       sizeof(char*) * (grid->colcount+1));
      grid->cells[i][grid->colcount] = NULL;
    }
    return grid->colcount++;
  }
  return -1;
}

void ag_remrow(a_grid_t* grid, int index) {
  if (grid && index >= 0 && index < grid->rowcount) {
    int i; for (i=0; i<grid->colcount; i++) {
      if (grid->cells[index][i] != NULL) free(grid->cells[index][i]);
    }
    free(grid->cells[index]);
    memmove(grid->cells + index, grid->cells + index + 1,
            sizeof(char***) * (grid->rowcount - index));
    grid->cells = (char***)realloc(grid->cells, sizeof(char**) *
                                  (grid->rowcount-1));
    (grid->rowcount)--;
  }
}

void ag_remcol(a_grid_t* grid, int index) {
  if (grid && index >= 0 && index < grid->colcount) {
    int i; for (i=0; i<grid->rowcount; i++) {
      if (grid->cells[i][index] != NULL) free(grid->cells[i][index]);
      memmove(grid->cells[i] + index, grid->cells[i] + index + 1,
              sizeof(char**) * (grid->colcount - index));
      grid->cells[i] = (char**)realloc(grid->cells[i], sizeof(char*) *
                                       (grid->colcount-1));
    }
    (grid->colcount)--;
  }
}

int ag_setrowcount(a_grid_t* grid, int cnt) {
  if (grid && cnt >= 0) {
    if (cnt > grid->rowcount) {
      int i, diff = cnt - grid->rowcount;
      for (i=0; i<diff; i++) {
        ag_addrow(grid);
      }
      return diff;
    } else if (cnt < grid->rowcount) {
      int i, diff = grid->rowcount - cnt;
      for (i=0; i<diff; i++) {
        ag_remrow(grid, grid->rowcount-1);
      }
      return -diff;
    } else return 0;
  }
  return 0;
}

int ag_setcolcount(a_grid_t* grid, int cnt) {
  if (grid && cnt >= 0) {
    if (cnt > grid->colcount) {
      int i, diff = cnt - grid->colcount;
      for (i=0; i<diff; i++) {
        ag_addcol(grid);
      }
      return diff;
    } else if (cnt < grid->colcount) {
      int i, diff = grid->colcount - cnt;
      for (i=0; i<diff; i++) {
        ag_remcol(grid, grid->colcount-1);
      }
      return -diff;
    } else return 0;
  }
  return 0;
}

void ag_insertrow(a_grid_t* grid, int index) {
  if (grid && index >= 0 && index < grid->rowcount) {
    int i;
    grid->cells = (char***)realloc(grid->cells, sizeof(char**) *
                                  (grid->rowcount+1));
    memmove(grid->cells + index + 1, grid->cells + index,
            grid->rowcount - index);
    grid->cells[index] = (char**)malloc(sizeof(char*) * grid->colcount);
    for (i=0; i<grid->colcount; i++) {
      grid->cells[index][i] = NULL;
    }
    (grid->rowcount)++;
  }
}

void ag_insertcol(a_grid_t* grid, int index) {
  if (grid && index >= 0 && index < grid->colcount) {
    int i; for (i=0; i<grid->rowcount; i++) {
      memmove(grid->cells[i] + index + 1, grid->cells[i] + index,
              grid->colcount - index);
      grid->cells[i] = (char**)realloc(grid->cells[i], sizeof(char*) *
                                       (grid->colcount+1));
      grid->cells[i][index] = NULL;
    }
    (grid->colcount)++;
  }
}

int ag_getrowcount(a_grid_t* grid) {
  return (grid ? grid->rowcount : -1);
}

int ag_getcolcount(a_grid_t* grid) {
  return (grid ? grid->colcount : -1);
}

void ag_set(a_grid_t* grid, int row, int col, char* value) {
  if (grid && row>=0 && col>=0) {
    while (row >= grid->rowcount) {
      ag_addrow(grid);
    }
    while (col >= grid->colcount) {
      ag_addcol(grid);
    }
    if (grid->cells[row][col] != NULL) {
      free(grid->cells[row][col]);
    }
    if (value) {
      grid->cells[row][col] = strdup(value);
    } else {
      grid->cells[row][col] = NULL;
    }
  }
}

char* ag_itoa(int n, char* buffer) {
  if (n==0) {
    strcpy(buffer, "0");
    return buffer;
  } else {
    char tmp[128], * tw = tmp;
    int i, l, n1 = n;
    if (n1 < 0) {
      n1 = -n1;
    }
    while (n1) {
      int d = (n1 - ((n1/10)*10));
      *tw++ = (char)(d + '0');
      n1 /= 10;
    }
    if (n < 0) *tw++ = '-';
    l = tw - tmp;
    for (i=0; i<l; i++) buffer[i] = tmp[l-i-1];
    buffer[i] = '\0';
    return buffer;
  }
}

void ag_setasint(a_grid_t* grid, int row, int col, int value) {
  if (grid && row>=0 && col>=0) {
    char buffer[128];
    while (row >= grid->rowcount) {
      ag_addrow(grid);
    }
    while (col >= grid->colcount) {
      ag_addcol(grid);
    }
    if (grid->cells[row][col] != NULL) {
      free(grid->cells[row][col]);
    }
    grid->cells[row][col] = strdup(ag_itoa(value, buffer));
  }
}

void ag_setastime(a_grid_t* grid, int row, int col, time_t value) {
  if (grid && row>=0 && col>=0) {
    char buffer[128];
    struct tm* tm = localtime(&value);
    while (row >= grid->rowcount) {
      ag_addrow(grid);
    }
    while (col >= grid->colcount) {
      ag_addcol(grid);
    }
    if (grid->cells[row][col] != NULL) {
      free(grid->cells[row][col]);
    }
    strftime(buffer, 128, "%d-%m-%Y %H:%M:%S", tm);
    grid->cells[row][col] = strdup(buffer);
  }
}

void ag_setasfloatwithprecision
  (a_grid_t* grid, int row, int col, double value, int p) {
  if (grid && row>=0 && col>=0) {
    char buffer[128];
    while (row >= grid->rowcount) {
      ag_addrow(grid);
    }
    while (col >= grid->colcount) {
      ag_addcol(grid);
    }
    if (grid->cells[row][col] != NULL) {
      free(grid->cells[row][col]);
    }
    snprintf(buffer, 128, "%.*f", p, value);
    grid->cells[row][col] = strdup(buffer);
  }
}

void ag_setasfloat(a_grid_t* grid, int row, int col, double value) {
  ag_setasfloatwithprecision(grid, row, col, value, 6);
}

char* ag_get(a_grid_t* grid, int row, int col) {
  if (grid &&
      row>=0 && row<grid->rowcount &&
      col>=0 && col<grid->colcount) {
    return grid->cells[row][col];
  }
  return NULL;
}

int ag_getasint(a_grid_t* grid, int row, int col) {
  if (grid &&
      row>=0 && row<grid->rowcount &&
      col>=0 && col<grid->colcount) {
    return ((grid->cells[row][col]) ? atoi(grid->cells[row][col]) : 0);
  }
  return 0;
}


time_t ag_getastime(a_grid_t* grid, int row, int col) {
  if (grid &&
      row>=0 && row<grid->rowcount &&
      col>=0 && col<grid->colcount &&
      grid->cells[row][col]) {
    struct tm tm;
    int Y = 0, m = 0, d = 0, H = 12, M = 0, S = 0;
    if (sscanf(grid->cells[row][col],
               "%d-%d-%d %d:%d:%d",
               &d, &m, &Y, &H, &M, &S) >= 3) {
      memset(&tm, 0, sizeof(tm));
      tm.tm_year  = Y - 1900;
      tm.tm_mon   = m - 1;
      tm.tm_mday  = d;
      tm.tm_hour  = H;
      tm.tm_min   = M;
      tm.tm_sec   = S;
      tm.tm_isdst = 1;
      return mktime(&tm);
    }
  }
  return (time_t)0;
}

double ag_getasfloat(a_grid_t* grid, int row, int col) {
  if (grid &&
      row>=0 && row<grid->rowcount &&
      col>=0 && col<grid->colcount &&
      grid->cells[row][col]) {
    return atof(grid->cells[row][col]);
  }
  return (double)0;
}

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

struct exportstretchbuffer {
  char* buffer;
  int offset;
  int length;
  int bufsiz;
};

char* ag_fgets(int fd, struct exportstretchbuffer* esb) {
  if (esb->offset) {
    memmove(esb->buffer, esb->buffer + esb->offset, esb->length - esb->offset);
    esb->length -= esb->offset;
    esb->offset = 0;
  }
  for (;;) {
    int n = read(fd, esb->buffer + esb->length, esb->bufsiz - esb->length);
    int i;
    esb->length += n;
    for (i=0; i<esb->length; i++) {
      if (esb->buffer[i] == '\\') {
        i++;
      } else if (esb->buffer[i] == '\n' || esb->buffer[i] == '\r') {
        esb->buffer[i] = '\0';
        esb->offset = i+1;
        return esb->buffer;
      }
    }
    if (!n) {
      if (!esb->length) {
        return NULL;
      } else {
        esb->buffer[i] = '\0';
//        esb->offset = i+1;
        esb->offset = i;
        return esb->buffer;
      }
    }
    esb->bufsiz += BUFSIZ;
    esb->buffer = (char*)realloc(esb->buffer, esb->bufsiz + 1);
    if (!esb->buffer) return NULL;
  }
}

char* ag_unescape(char* string) {
  char* s0 = string, * s1 = string;
  while (*s1) {
    if (*s1 == '\\') {
      ++s1;
    }
    *s0++ = *s1++;
  }
  *s0 = '\0';
  return string;
}

char* buffer2cell(char* offset, a_grid_t* grid, int row, int col) {
  char* o = offset;
  for (;;) {
    if (*o == '\\') {
      ++o;
    } else if (*o == '\0' || *o == '/') {
      char orig = *o;
      *o = '\0';
      if (!strcmp(offset, "(null)")) {
        ag_set(grid, row, col, NULL);
      } else {
        ag_set(grid, row, col, ag_unescape(offset));
      }
      if (!orig) return NULL;
      return ++o;
    }
    ++o;
  }
}

int ag_fread(FILE* f, a_grid_t* grid) {
  return ag_fdread(fileno(f), grid);
/*
  if (f && grid) {
    struct exportstretchbuffer esb;
    int row = 0, col = 0;
    esb.buffer = (char*)malloc(BUFSIZ + 1);
    esb.bufsiz = BUFSIZ;
    esb.offset = 0;
    esb.length = 0;
    while (ag_fgets(f, &esb)) {
      char* offset = esb.buffer;
      col = 0;
      if (esb.buffer[0] == '#' || esb.buffer[0] == '\0') continue;
      while ((offset = buffer2cell(offset, grid, row, col))) {
        col++;
      }
      row++;
    }
    free(esb.buffer);
  }
  return -1;
*/
}

int ag_fdread(int fd, a_grid_t* grid) {
  if (fd >= 0 && grid) {
    struct exportstretchbuffer esb;
    int row = 0, col = 0;
    esb.buffer = (char*)malloc(BUFSIZ + 1);
    esb.bufsiz = BUFSIZ;
    esb.offset = 0;
    esb.length = 0;
    while (ag_fgets(fd, &esb)) {
      char* offset = esb.buffer;
      col = 0;
      if (esb.buffer[0] == '#' || esb.buffer[0] == '\0') continue;
      while ((offset = buffer2cell(offset, grid, row, col))) {
        col++;
      }
      row++;
    }
    free(esb.buffer);
  }
  return -1;
}

int ag_read(char* path, a_grid_t* grid) {
  int fd;
#ifdef WIN32
  _fmode = _O_BINARY;
#endif
  fd = open(path, O_RDONLY);
  if (fd >= 0) {
    int result = ag_fdread(fd, grid);
    close(fd);
    return result;
  }
  return -1;
}

char* ag_mallocescape(char* string) {
  if (string) {
    char* result = (char*)malloc(strlen(string)*2 + 1), * r = result;
    if (!result) return NULL;
    while (*string) {
      if (strchr("\\\n/()#", *string)) {
        *r++ = '\\';
      }
      *r++ = *string++;
    }
    *r = '\0';
    return result;
  } else {
    return strdup("(null)");
  }
}

int ag_fwrite(FILE* f, a_grid_t* grid) {
  return ag_fdwrite(fileno(f), grid);
/*
  if (f && grid) {
    int i;
    fprintf(f, "# -- a_grid version %d.%d --\n",
               AG_VERSION_MAJOR, AG_VERSION_MINOR);
    for (i=0; i<grid->rowcount; i++) {
      int j;
      for (j=0; j<grid->colcount; j++) {
        char* esc = ag_mallocescape(grid->cells[i][j]);
        fprintf(f, "%s%s", esc, ((j==grid->colcount-1) ? "" : "/"));
        free(esc);
      }
      fputc('\n', f);
    }
  }
  return -1;
*/
}

int ag_fdwrite(int fd, a_grid_t* grid) {
  if (fd >= 0 && grid) {
    int i;
    char header[BUFSIZ];
    snprintf(header, BUFSIZ, "# -- a_grid version %d.%d --\n",
                             AG_VERSION_MAJOR, AG_VERSION_MINOR);
    write(fd, header, strlen(header));
    for (i=0; i<grid->rowcount; i++) {
      int j;
      for (j=0; j<grid->colcount; j++) {
        char* esc = ag_mallocescape(grid->cells[i][j]);
        if (!esc) {
          write(fd, "(too big)", 9);
        } else {
          write(fd, esc, strlen(esc));
          free(esc);
        }
        if (j < grid->colcount-1) write(fd, "/", 1);
      }
      write(fd, "\n", 1);
    }
  }
  return -1;
}

int ag_write(char* path, a_grid_t* grid) {
  int fd;
#ifdef WIN32
  _fmode = _O_BINARY;
#endif
  fd = open(path, O_WRONLY|O_CREAT|O_TRUNC, 0664);
  if (fd >= 0) {
    int result = ag_fdwrite(fd, grid);
    close(fd);
    return result;
  }
  return -1;
}


//---------- named grids ------------
//---------- warning: no ANSI-C !

void pg_pushrelicpointer(a_ngrid_t* ngrid, void* mem) {
  ngrid->relic = (void**)realloc(ngrid->relic,
                                 sizeof(void*) * (ngrid->reliccount+1));
  ngrid->relic[ngrid->reliccount++] = mem;
}

void pg_examinepointer(a_ngrid_t* ngrid, int row, int col) {
  if (row >= 0 && row < ngrid->rowcount &&
      col >= 0 && col < ngrid->colcount &&
      ngrid->cells[row][col] != NULL) {
    pg_pushrelicpointer(ngrid, ngrid->cells[row][col]);
    ngrid->cells[row][col] = NULL;
  }
}

void pg_set(a_ngrid_t* ngrid, int row, int col, char* value) {
  pg_examinepointer(ngrid, row, col);
  ag_set((a_grid_t*)ngrid, row, col, value);
  ngrid->rowselect = 0;
  ngrid->colselect = 0;
  ngrid->changed   = 1;
}

void pg_setasint(a_ngrid_t* ngrid, int row, int col, int value) {
  pg_examinepointer(ngrid, row, col);
  ag_setasint((a_grid_t*)ngrid, row, col, value);
  ngrid->rowselect = 0;
  ngrid->colselect = 0;
  ngrid->changed   = 1;
}

void pg_setastime(a_ngrid_t* ngrid, int row, int col, time_t value) {
  pg_examinepointer(ngrid, row, col);
  ag_setastime((a_grid_t*)ngrid, row, col, value);
  ngrid->rowselect = 0;
  ngrid->colselect = 0;
  ngrid->changed   = 1;
}

void pg_setasfloat(a_ngrid_t* ngrid, int row, int col, double value) {
  pg_examinepointer(ngrid, row, col);
  ag_setasfloatwithprecision((a_grid_t*)ngrid, row, col,
                             value, ngrid->double_precision);
  ngrid->rowselect = 0;
  ngrid->colselect = 0;
  ngrid->changed   = 1;
}

a_ngrid_t* g_init() {
  a_ngrid_t* result;
  result = (a_ngrid_t*)malloc(sizeof(a_ngrid_t));
  if (result) {
    result->cells            = NULL;
    result->rowcount         = 0;
    result->colcount         = 0;
    result->rowselect        = 0;
    result->colselect        = 0;
    result->fd               = -1;
    result->changed          = 0;
    result->double_precision = 6;
    result->relic            = NULL;
    result->reliccount       = 0;
  }
  return result;
}

a_ngrid_t* g_open(char* path) {
  if (path) {
    a_ngrid_t* result;
    int fd;
#ifdef WIN32
    _fmode = _O_BINARY;
#endif
    fd = open(path, O_CREAT|O_RDWR, 0664);
    if (fd >= 0) {
      result = (a_ngrid_t*)malloc(sizeof(a_ngrid_t));
      if (result) {
        result->cells            = NULL;
        result->rowcount         = 0;
        result->colcount         = 0;
        result->rowselect        = 0;
        result->colselect        = 0;
        result->fd               = fd;
        result->changed          = 0;
        result->double_precision = 6;
        result->relic            = NULL;
        result->reliccount       = 0;
        ag_fdread(fd, (a_grid_t*)result);
      }
      return result;
    }
  }
  return NULL;
}

void g_save(a_ngrid_t* ngrid) {
  if (ngrid) {
    if (ngrid->changed && ngrid->fd >= 0) {
      ftruncate(ngrid->fd, 0);
      lseek(ngrid->fd, 0, SEEK_SET);
      ag_fdwrite(ngrid->fd, (a_grid_t*)ngrid);
      ngrid->changed = 0;
    }
  }
}

void g_saveas(a_ngrid_t* ngrid, char* path) {
  if (ngrid) {
    int fd;
#ifdef WIN32
    _fmode = _O_BINARY;
#endif
    fd = open(path, O_CREAT|O_RDWR, 0664);
    if (fd >= 0) {
      if (ngrid->fd >= 0) close(ngrid->fd);
      ngrid->fd = fd;
      ngrid->changed = 1;
      g_save(ngrid);
    }
  }
}

void g_flush(a_ngrid_t* ngrid) {
  if (ngrid) {
    if (ngrid->reliccount || ngrid->relic) {
      int i;
      for (i=0; i<ngrid->reliccount; i++) {
        if (ngrid->relic[i] != NULL) free(ngrid->relic[i]);
      }
      free(ngrid->relic);
      ngrid->relic = NULL;
      ngrid->reliccount = 0;
    }
  }
}

void g_sync(a_ngrid_t* ngrid) {
  g_flush(ngrid);
  g_save(ngrid);
}

void g_close(a_ngrid_t* ngrid) {
  if (ngrid) {
    g_sync(ngrid);
    if (ngrid->fd >= 0) close(ngrid->fd);
    ag_deletegrid((a_grid_t*)ngrid);
  }
}

void g_setfloatprecision(a_ngrid_t* ngrid, int p) {
  if (ngrid && p >= 0 && p < 20) {
    ngrid->double_precision = p;
  }
}

void g_settop(a_ngrid_t* ngrid, char* value) {
  if (ngrid) {
    pg_set(ngrid, 0, 0, value);
  }
}

char* g_gettop(a_ngrid_t* ngrid) {
  if (ngrid) {
    return ag_get((a_grid_t*)ngrid, 0, 0);
  }
  return NULL;
}

int g_getrowcount(a_ngrid_t* ngrid) {
  return (ag_getrowcount((a_grid_t*)ngrid));
}

int g_getcolcount(a_ngrid_t* ngrid) {
  return (ag_getcolcount((a_grid_t*)ngrid));
}

int g_addrow(a_ngrid_t* ngrid, char* value) {
  if (ngrid && ngrid->rowcount && ngrid->colcount) {
    int index = ag_addrow((a_grid_t*)ngrid);
    pg_set(ngrid, index, 0, value);
    return index;
  }
  return -1;
}

int g_addcol(a_ngrid_t* ngrid, char* value) {
  if (ngrid && ngrid->rowcount && ngrid->colcount) {
    int index = ag_addcol((a_grid_t*)ngrid);
    pg_set(ngrid, 0, index, value);
    return index;
  }
  return -1;
}

int g_addrowasint(a_ngrid_t* ngrid, int value) {
  if (ngrid && ngrid->rowcount && ngrid->colcount) {
    int index = ag_addrow((a_grid_t*)ngrid);
    pg_setasint(ngrid, index, 0, value);
    return index;
  }
  return -1;
}

int g_addcolasint(a_ngrid_t* ngrid, int value) {
  if (ngrid && ngrid->rowcount && ngrid->colcount) {
    int index = ag_addcol((a_grid_t*)ngrid);
    pg_setasint(ngrid, 0, index, value);
    return index;
  }
  return -1;
}

void g_remrowindex(a_ngrid_t* grid, int index) {
  ag_remrow((a_grid_t*)grid, index);
  grid->changed = 1;
}

void g_remcolindex(a_ngrid_t* grid, int index) {
  ag_remcol((a_grid_t*)grid, index);
  grid->changed = 1;
}

int g_findinrow(a_ngrid_t* ngrid, int row, char* value) {
  if (ngrid && row >= 0 && row < ngrid->rowcount && value) {
    int i;
    if (++(ngrid->colselect) >= ngrid->colcount) ngrid->colselect = 0;
    for (i=ngrid->colselect; i<ngrid->colcount; i++) {
      if (!strcmp(ngrid->cells[row][i], value)) {
        ngrid->colselect = i;
        return i;
      }
    }
    for (i=0; i<ngrid->colselect; i++) {
      if (!strcmp(ngrid->cells[row][i], value)) {
        ngrid->colselect = i;
        return i;
      }
    }
  }
  return -1;
}

int g_findincol(a_ngrid_t* ngrid, int col, char* value) {
  if (ngrid && col >= 0 && col < ngrid->colcount && value) {
    int i;
    if (++(ngrid->rowselect) >= ngrid->rowcount) ngrid->rowselect = 0;
    for (i=ngrid->rowselect; i<ngrid->rowcount; i++) {
      if (ngrid->cells[i][col] && !strcmp(ngrid->cells[i][col], value)) {
        ngrid->rowselect = i;
        return i;
      }
    }
    for (i=0; i<ngrid->rowselect; i++) {
      if (ngrid->cells[i][col] && !strcmp(ngrid->cells[i][col], value)) {
        ngrid->rowselect = i;
        return i;
      }
    }
  }
  return -1;
}

int g_getrowindex(a_ngrid_t* ngrid, char* value) {
  if (ngrid && value) {
    return g_findincol(ngrid, 0, value);
  }
  return -1;
}

int g_hasrow(a_ngrid_t* ngrid, char* rowvalue) {
  return (g_getrowindex(ngrid, rowvalue) != -1);
}

int g_getcolindex(a_ngrid_t* ngrid, char* value) {
  if (ngrid && value) {
    return g_findinrow(ngrid, 0, value);
  }
  return -1;
}

int g_hascol(a_ngrid_t* ngrid, char* colvalue) {
  return (g_getcolindex(ngrid, colvalue) != -1);
}

int g_findintincol(a_ngrid_t* ngrid, int col, int value) {
  if (ngrid && col >= 0 && col < ngrid->colcount) {
    int i;
    if (++(ngrid->rowselect) >= ngrid->rowcount) ngrid->rowselect = 0;
    for (i=ngrid->rowselect; i<ngrid->rowcount; i++) {
      if (atoi(ngrid->cells[i][col]) == value) {
        ngrid->rowselect = i;
        return i;
      }
    }
    for (i=0; i<ngrid->rowselect; i++) {
      if (atoi(ngrid->cells[i][col]) == value) {
        ngrid->rowselect = i;
        return i;
      }
    }
  }
  return -1;
}


int g_findintinrow(a_ngrid_t* ngrid, int row, int value) {
  if (ngrid && row >= 0 && row < ngrid->rowcount) {
    int i;
    if (++(ngrid->colselect) >= ngrid->colcount) ngrid->colselect = 0;
    for (i=ngrid->colselect; i<ngrid->colcount; i++) {
      if (atoi(ngrid->cells[row][i]) == value) {
        ngrid->colselect = i;
        return i;
      }
    }
    for (i=0; i<ngrid->colselect; i++) {
      if (atoi(ngrid->cells[row][i]) == value) {
        ngrid->colselect = i;
        return i;
      }
    }
  }
  return -1;
}

int g_getrowindexasint(a_ngrid_t* ngrid, int value) {
  return g_findintincol(ngrid, 0, value);
}


int g_getcolindexasint(a_ngrid_t* ngrid, int value) {
  return g_findintinrow(ngrid, 0, value);
}

void g_set(a_ngrid_t* ngrid, char* rowvalue, char* colvalue, char* value) {
  if (ngrid && rowvalue && colvalue) {
    int row = g_getrowindex(ngrid, rowvalue);
    int col = g_getcolindex(ngrid, colvalue);
    if (row >= 0 && col >= 0) {
      pg_set(ngrid, row, col, value);
    }
  }
}

void g_setasint(a_ngrid_t* ngrid, char* rowvalue, char* colvalue, int value) {
  if (ngrid) {
    char numberstring[128];
    ag_itoa(value, numberstring);
    g_set(ngrid, rowvalue, colvalue, numberstring);
  }
}

char* g_get(a_ngrid_t* ngrid, char* rowvalue, char* colvalue) {
  if (ngrid && rowvalue && colvalue) {
    int row = g_getrowindex(ngrid, rowvalue);
    int col = g_getcolindex(ngrid, colvalue);
    if (row >= 0 && col >= 0) {
      return ag_get((a_grid_t*)ngrid, row, col);
    }
  }
  return NULL;
}

int g_getasint(a_ngrid_t* ngrid, char* rowvalue, char* colvalue) {
  if (ngrid && rowvalue && colvalue) {
    int row = g_getrowindex(ngrid, rowvalue);
    int col = g_getcolindex(ngrid, colvalue);
    if (row >= 0 && col >= 0) {
      return atoi(ag_get((a_grid_t*)ngrid, row, col));
    }
  }
  return 0;
}

char* g_getatrow(a_ngrid_t* ngrid, char* colvalue) {
  if (ngrid && colvalue) {
    int col = g_getcolindex(ngrid, colvalue);
    if (ngrid->rowselect >= 0 && col >= 0) {
      return ag_get((a_grid_t*)ngrid, ngrid->rowselect, col);
    }
  }
  return NULL;
}

int g_getatrowasint(a_ngrid_t* ngrid, char* colvalue) {
  if (ngrid && colvalue) {
    int col = g_getcolindex(ngrid, colvalue);
    if (ngrid->rowselect >= 0 && col >= 0) {
      return atoi(ag_get((a_grid_t*)ngrid, ngrid->rowselect, col));
    }
  }
  return 0;
}

char* g_getcell(a_ngrid_t* ngrid, int row, int col) {
  return ag_get((a_grid_t*)ngrid, row, col);
}

int g_getcellasint(a_ngrid_t* ngrid, int row, int col) {
  return ag_getasint((a_grid_t*)ngrid, row, col);
}

time_t g_getcellastime(a_ngrid_t* ngrid, int row, int col) {
  return ag_getastime((a_grid_t*)ngrid, row, col);
}

double g_getcellasfloat(a_ngrid_t* ngrid, int row, int col) {
  return ag_getasfloat((a_grid_t*)ngrid, row, col);
}

int g_cellisset(a_ngrid_t* ngrid, int row, int col) {
  return (g_getcell(ngrid, row, col) ? 1 : 0);
}

void g_setcell(a_ngrid_t* ngrid, int row, int col, char* value) {
  if (ngrid) {
    pg_set(ngrid, row, col, value);
  }
}

int g_rownset(a_ngrid_t* ngrid, int row, char** values, int size) {
  if (ngrid) {
    int i;
    for (i=0; i<size; i++) {
      pg_set(ngrid, row, i, values[i]);
    }
    return 0;
  }
  return -1;
}

int g_colnset(a_ngrid_t* ngrid, int col, char** values, int size) {
  if (ngrid) {
    int i;
    for (i=0; i<size; i++) {
      pg_set(ngrid, i, col, values[i]);
    }
    return 0;
  }
  return -1;
}

void g_setcellasint(a_ngrid_t* ngrid, int row, int col, int value) {
  if (ngrid) {
    pg_setasint(ngrid, row, col, value);
  }
}

void g_setcellastime(a_ngrid_t* ngrid, int row, int col, time_t value) {
  if (ngrid) {
    pg_setastime(ngrid, row, col, value);
  }
}

void g_setcellasfloat(a_ngrid_t* ngrid, int row, int col, double value) {
  if (ngrid) {
    pg_setasfloat(ngrid, row, col, value);
  }
}

void g_remrow(a_ngrid_t* ngrid, char* value) {
  int index = g_getrowindex(ngrid, value);
  if (index != -1) {
    ag_remrow((a_grid_t*)ngrid, index);
    ngrid->changed = 1;
  }
}

void g_remcol(a_ngrid_t* ngrid, char* value) {
  int index = g_getcolindex(ngrid, value);
  if (index != -1) {
    ag_remcol((a_grid_t*)ngrid, index);
    ngrid->changed = 1;
  }
}

int g_setrowcount(a_ngrid_t* ngrid, int cnt) {
  if (ngrid && cnt >= 0) {
    int result = ag_setrowcount((a_grid_t*)ngrid, cnt);
    if (result) ngrid->changed = 1;
    return result;
  }
  return 0;
}

int g_setcolcount(a_ngrid_t* ngrid, int cnt) {
  if (ngrid && cnt >= 0) {
    int result = ag_setcolcount((a_grid_t*)ngrid, cnt);
    if (result) ngrid->changed = 1;
    return result;
  }
  return 0;
}

int g_rowcount(a_ngrid_t* ngrid, char* colname, char* cell) {
  if (ngrid && colname) {
    int colindex = g_findinrow(ngrid, 0, colname);
    if (colindex != -1) {
      int result = 0;
      int i; for (i=1; i<ngrid->rowcount; i++) {
        if (cell == NULL) {
          if (ngrid->cells[i][colindex] == NULL) result++;
        } else {
          if (ngrid->cells[i][colindex] != NULL &&
              !strcmp(ngrid->cells[i][colindex], cell)) result++;
        }
      }
      return result;
    }
  }
  return -1;
}

int pg_colncmp(a_ngrid_t* ngrid, int col, char** cols, int size, int (*cmp)()) {
  if (ngrid && col >= 0 && col < ngrid->colcount && cols) {
    int i; for (i=0; i<size; i++) {
      if (ngrid->cells[i][col]) {
        if (cols[i]) {
          int c = cmp(ngrid->cells[i][col], cols[i]);
          if (c) return c;
        } else {
          return 1;
        }
      } else {
        if (cols[i]) {
          return -1;
        }
      }
    }
    return 0;
  }
  return -1;
}


int g_colncmp(a_ngrid_t* ngrid, int col, char** cols, int size) {
  return pg_colncmp(ngrid, col, cols, size, strcmp);
}

int g_colncasecmp(a_ngrid_t* ngrid, int col, char** cols, int size) {
  return pg_colncmp(ngrid, col, cols, size, strcasecmp);
}

int pg_rowncmp(a_ngrid_t* ngrid, int row, char** rows, int size, int (*cmp)()) {
  if (ngrid && row >= 0 && row < ngrid->rowcount && rows) {
    int i; for (i=0; i<size; i++) {
      if (ngrid->cells[row][i]) {
        if (rows[i]) {
          int c = cmp(ngrid->cells[row][i], rows[i]);
          if (c) return c;
        } else {
          return 1;
        }
      } else {
        if (rows[i]) {
          return -1;
        }
      }
    }
    return 0;
  }
  return -1;
}

int g_rowncmp(a_ngrid_t* ngrid, int row, char** rows, int size) {
  return pg_rowncmp(ngrid, row, rows, size, strcmp);
}

int g_rowncasecmp(a_ngrid_t* ngrid, int row, char** rows, int size) {
  return pg_rowncmp(ngrid, row, rows, size, strcasecmp);
}

int g_cellcmp(a_ngrid_t* ngrid, int row1, int col1, int row2, int col2,
              int (*cmp)()) {
  if (ngrid && row1 >= 0 && row1 < ngrid->rowcount &&
               col1 >= 0 && col1 < ngrid->colcount &&
               row2 >= 0 && row2 < ngrid->rowcount &&
               col2 >= 0 && col2 < ngrid->colcount) {
    if (ngrid->cells[row1][col1]) {
      if (ngrid->cells[row2][col2]) {
        return cmp(ngrid->cells[row1][col1], ngrid->cells[row2][col2]);
      } else {
        return 1;
      }
    } else if (ngrid->cells[row2][col2]) {
      return -1;
    }
  }
  return 0;
}

int g_colswap(a_ngrid_t* ngrid, int col1, int col2) {
  if (ngrid && col1 >= 0 && col1 < ngrid->colcount &&
               col2 >= 0 && col2 < ngrid->colcount) {
    int i;
    for (i=0; i<ngrid->rowcount; i++) {
      char* tmp = ngrid->cells[i][col1];
      ngrid->cells[i][col1] = ngrid->cells[i][col2];
      ngrid->cells[i][col2] = tmp;
    }
    return 0;
  }
  return -1;
}

int g_rowswap(a_ngrid_t* ngrid, int row1, int row2) {
  if (ngrid && row1 >= 0 && row1 < ngrid->rowcount &&
               row2 >= 0 && row2 < ngrid->rowcount) {
    int i;
    for (i=0; i<ngrid->colcount; i++) {
      char* tmp = ngrid->cells[row1][i];
      ngrid->cells[row1][i] = ngrid->cells[row2][i];
      ngrid->cells[row2][i] = tmp;
    }
    return 0;
  }
  return -1;
}

int g_sortrow(a_ngrid_t* ngrid, int row, int flags) {
  if (ngrid && row >= 0 && row < ngrid->rowcount) {
    int done = 0;
    int (*cmp)() = strcmp;
    if (flags & SORT_CASE_INSENSITIVE) cmp = strcasecmp;
    while (!done) {
      int i;
      done = 1;
      for (i=0; i<ngrid->colcount-1; i++) {
        int a = g_cellcmp(ngrid, row, i, row, i+1, cmp);
        if (((flags & SORT_DESCENDING) && a < 0) ||
            (!(flags & SORT_DESCENDING) && a > 0)) {
          g_colswap(ngrid, i, i+1);
          done = 0;
          break;
        }
      }
    }
  }
  return -1;
}

int g_sortcol(a_ngrid_t* ngrid, int col, int flags) {
  if (ngrid && col >= 0 && col < ngrid->colcount) {
    int done = 0;
    int (*cmp)() = strcmp;
    if (flags & SORT_CASE_INSENSITIVE) cmp = strcasecmp;
    while (!done) {
      int i;
      done = 1;
      for (i=0; i<ngrid->rowcount-1; i++) {
        int a = g_cellcmp(ngrid, i, col, i+1, col, cmp);
        if (((flags & SORT_DESCENDING) && a < 0) ||
            (!(flags & SORT_DESCENDING) && a > 0)) {
          g_rowswap(ngrid, i, i+1);
          done = 0;
          break;
        }
      }
    }
  }
  return -1;
}


a_ngrid_t* g_dup_rect(a_ngrid_t* ngrid, int top, int left,
                      int height, int width) {
  if (ngrid && top >= 0 && top + height < ngrid->rowcount &&
      left >= 0 && left + width < ngrid->colcount) {
    a_ngrid_t* result = g_init();
    if (result) {
      int i; for (i=0; i<ngrid->rowcount; i++) {
        int j; for (j=0; j<ngrid->colcount; j++) {
          g_setcell(result, i, j, g_getcell(ngrid, i, j));
        }
      }
    }
    return result;
  }
  return NULL;
}

a_ngrid_t* g_dup(a_ngrid_t* ngrid) {
  return g_dup_rect(ngrid, 0, 0, ngrid->rowcount, ngrid->colcount);
}

int g_merge(a_ngrid_t* dst, a_ngrid_t* src) {
  //.. merge src into dst where rows and cols match
  return -1;
}

int g_copyrow_add(a_ngrid_t* dst, a_ngrid_t* src, int row) {
  if (dst && src && row >= 0 && row < src->rowcount) {
    int dstrow = dst->rowcount;
    int i; for (i=0; i<src->colcount; i++) {
      g_setcell(dst, dstrow, i, g_getcell(src, row, i));
    }
    return 0;
  }
  return -1;
}

char* g_strdup(a_ngrid_t* g) {
  return NULL;
}


/*
 * Compiling this code with -D__A_GRID_MAIN__ (or something that sets
 * the following #definition), will result in an executable for examining
 * the file format that this api generates.
 */

#ifdef __A_GRID_MAIN__

void writeuse(char* appname) {
  printf("Usage: %s [options] filename (action [parameters])\n"
         "options: preceeded by '-', combine the following letters:\n"
         "         r set row:\n"
         "         c set column:\n"
         "         u unique (no more the same rows/columns allowed)\n"
         "         d drop (the same rows/columns will replace the old ones)\n"
         "         n non-unique (the same rows/columns will be added)\n"
         "actions: (mutating actions - no output)\n"
         "         'addrow' rowname\n"
         "         'addcol' columnname\n"
         "         'setcell' rowname columnname cellvalue\n"
         "         (actions that generate output over stdout)\n"
         "         'getcell' rowname columnname\n"
         "         'show' (same as no action)\n", appname);
}

void showgrid(char* filename, int options) {
  int i, j;
  a_grid_t* grid = ag_creategrid(options);
  ag_read(filename, grid);
  for (j=0; j<grid->rowcount; j++) {
    for (i=0; i<grid->colcount; i++) {
      printf("[%s]%s", grid->cells[j][i],
             ((i==grid->colcount-1) ? "" : "\t"));
    }
    printf("\n");
  }
  ag_deletegrid(grid);
}

int main(int argc, char** argv) {
  int options = 0, i = 1;
  FILE* f;
  if (argc == 1) {
    writeuse(argv[0]); exit(-1);
  }
  // filter out the options
  for (;;) {
    if (*argv[i] == '-') {
      int setrowcol = 0;
      char* opt = argv[i]+1;
      while (*opt) {
        switch (*opt) {
        case 'r':
          setrowcol = 0;
          break;
        case 'c':
          setrowcol = 1;
          break;
        case 'u':
          if (setrowcol) options |= COLS_UNIQUE;
          else options |= ROWS_UNIQUE;
          break;
        case 'd':
          if (setrowcol) options |= COLS_UNIQUE_DROP;
          else options |= ROWS_UNIQUE_DROP;
          break;
        case 'n':
          if (setrowcol) options |= COLS_NON_UNIQUE;
          else options |= ROWS_NON_UNIQUE;
          break;
        }
        opt++;
      }
      if (++i >= argc) {
        writeuse(argv[0]);
        exit(-1);
      }
    } else break;
  }
  // test path
  f = fopen(argv[i], "r");
  if (f) {
    fclose(f);
  } else {
    printf("Could not open file %s.\n", argv[i]);
    writeuse(argv[0]); exit(-1);
  }
  // test action
  if (++i >= argc || !strcmp(argv[i], "show")) {
    showgrid(argv[i-1], options);
    exit(0);
  } else if (!strcmp(argv[i], "addrow")) {
    a_ngrid_t* ngrid;
    // expecting at least 1 more arguments
    if (i >= argc - 1) {
      writeuse(argv[0]);
      exit(-1);
    }
    ngrid = g_open(argv[i-1]);
    g_addrow(ngrid, argv[i+1]);
    g_close(ngrid);
  } else if (!strcmp(argv[i], "addcol")) {
    a_ngrid_t* ngrid;
    // expecting at least 1 more arguments
    if (i >= argc - 1) {
      writeuse(argv[0]);
      exit(-1);
    }
    ngrid = g_open(argv[i-1]);
    g_addcol(ngrid, argv[i+1]);
    g_close(ngrid);
  } else if (!strcmp(argv[i], "setcell")) {
    a_ngrid_t* ngrid;
    // expecting at least 3 more arguments
    if (i >= argc - 3) {
      writeuse(argv[0]);
      exit(-1);
    }
    ngrid = g_open(argv[i-1]);
    g_set(ngrid, argv[i+1], argv[i+2], argv[i+3]);
    g_close(ngrid);
  } else if (!strcmp(argv[i], "getcell")) {
    a_ngrid_t* ngrid;
    // expecting at least two more arguments
    if (i >= argc - 2) {
      writeuse(argv[0]);
      exit(-1);
    }
    ngrid = g_open(argv[i-1]);
    printf("%s", g_get(ngrid, argv[i+1], argv[i+2]));
//    g_close(ngrid);
  }
  return 0;
}

#endif

#ifdef __cplusplus
}
#endif
