/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * $Id: a_grid.h,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#ifndef __A_GRID_H__

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifdef __cplusplus
extern "C" {
#endif

#define AG_VERSION_MAJOR 1
#define AG_VERSION_MINOR 0

/*
 *                                                          05-05-2000
 *                                  This is the headerfile of a_grid.c
 * and some more derived types.
 * These functions read from disk, modify in memory, and write to disk,
 * a two dimensional grid of strings.
 * The file-format it uses is very human and machine readable - check it out.
 * For now, it is very elementary; I plan to add adding and subtracting-
 * capability of two grids, as well as the taking up of more data types
 * (converting them to strings internally) - ints, blobs (base64 encoded),
 * dates, etcetera. (and maybe some regexp or calculations ?)
 *
 */

typedef struct  {
  char*** cells;
  int     rowcount;
  int     colcount;
} a_grid_t;

a_grid_t* ag_creategrid();
void      ag_deletegrid(a_grid_t*);

int ag_getrowcount(a_grid_t*);
int ag_getcolcount(a_grid_t*);

int ag_setrowcount(a_grid_t*, int cnt);
int ag_setcolcount(a_grid_t*, int cnt);

int ag_addrow(a_grid_t*);
int ag_addcol(a_grid_t*);

void ag_remrow(a_grid_t*, int index);
void ag_remcol(a_grid_t*, int index);

void ag_set(a_grid_t*, int row, int col, char* value);
void ag_setasint(a_grid_t*, int row, int col, int value);
void ag_setastime(a_grid_t*, int row, int col, time_t value);
void ag_setasfloat(a_grid_t*, int row, int col, double value);

char*  ag_get(a_grid_t*, int row, int col);
int    ag_getasint(a_grid_t*, int row, int col);
time_t ag_getastime(a_grid_t*, int row, int col);
double ag_getasfloat(a_grid_t*, int row, int col);

int ag_read(char* path, a_grid_t*);
int ag_write(char* path, a_grid_t*);

int ag_fdread(int fd, a_grid_t* grid);
int ag_fdwrite(int fd, a_grid_t* grid);

int ag_fread(FILE* file, a_grid_t*);
int ag_fwrite(FILE* file, a_grid_t*);


//-- the row-/col-value abstraction layer

/*
 * A 'named grid' is a grid whose first row and first column are used for
 * 'naming' them.  It is used much in a way open() works; as a file-pointer.
 * However, one can also use them in memory only; and what's more; it will
 * remember all it's discarded pointers (So with many manipulations the
 * memory it uses grows) until you can g_sync() or g_close()
 */

typedef struct {
  char*** cells;
  int     rowcount;
  int     colcount;
  int     rowselect;
  int     colselect;
  int     fd;
  int     changed;
  int     double_precision;
  void**  relic;
  int     reliccount;
} a_ngrid_t;

a_ngrid_t* g_init();
a_ngrid_t* g_open(char* filename);
void       g_save(a_ngrid_t* ngrid);
void       g_saveas(a_ngrid_t* ngrid, char* filename);
void       g_flush(a_ngrid_t* ngrid); // free irrelevant pointers
void       g_sync(a_ngrid_t* ngrid);  // g_flush + g_save
void       g_close(a_ngrid_t* ngrid);

void g_setfloatprecision(a_ngrid_t*, int p);

int g_getrowcount(a_ngrid_t*);
int g_getcolcount(a_ngrid_t*);

int g_setrowcount(a_ngrid_t*, int cnt);
int g_setcolcount(a_ngrid_t*, int cnt);

int g_findinrow(a_ngrid_t* ngrid, int row, char* value);
int g_findincol(a_ngrid_t* ngrid, int col, char* value);
int g_findintinrow(a_ngrid_t* ngrid, int row, int value);
int g_findintincol(a_ngrid_t* ngrid, int col, int value);

int g_getrowindex(a_ngrid_t*, char* value);
int g_getcolindex(a_ngrid_t*, char* value);

int g_getrowindexasint(a_ngrid_t*, int value);
int g_getcolindexasint(a_ngrid_t*, int value);

int g_addrow(a_ngrid_t*, char* value);
int g_addcol(a_ngrid_t*, char* value);

void g_remrowindex(a_ngrid_t*, int index);
void g_remcolindex(a_ngrid_t*, int index);

void g_remrow(a_ngrid_t*, char* value);
void g_remcol(a_ngrid_t*, char* value);

int g_addrowasint(a_ngrid_t*, int value);
int g_addcolasint(a_ngrid_t*, int value);

void  g_setrow(a_ngrid_t*, char* value);
void  g_setcol(a_ngrid_t*, char* value);
char* g_getrowvalue(a_ngrid_t*, int index);
char* g_getcolvalue(a_ngrid_t*, int index);

int g_hasrow(a_ngrid_t*, char* value);
int g_hascol(a_ngrid_t*, char* value);

void  g_set(a_ngrid_t*, char* rowvalue, char* colvalue, char* value);
char* g_get(a_ngrid_t*, char* rowvalue, char* colvalue);

void g_setasint(a_ngrid_t*, char* rowvalue, char* colvalue, int value);
int  g_getasint(a_ngrid_t*, char* rowvalue, char* colvalue);

char* g_getatrow(a_ngrid_t* ngrid, char* colvalue);
int g_getatrowasint(a_ngrid_t* ngrid, char* colvalue);

void g_setcell(a_ngrid_t*, int row, int col, char* value);
void g_setcellasint(a_ngrid_t*, int row, int col, int value);
void g_setcellastime(a_ngrid_t*, int row, int col, time_t value);
void g_setcellasfloat(a_ngrid_t*, int row, int col, double value);

int g_cellisset(a_ngrid_t*, int row, int col);

char*  g_getcell(a_ngrid_t*, int row, int col);
int    g_getcellasint(a_ngrid_t*, int row, int col);
time_t g_getcellastime(a_ngrid_t*, int row, int col);
double g_getcellasfloat(a_ngrid_t*, int row, int col);

int g_rowcount(a_ngrid_t*, char* colname, char* cell);

int g_colncmp(a_ngrid_t*, int col, char** cols, int size);
int g_colncasecmp(a_ngrid_t*, int col, char** cols, int size);
int g_rowncmp(a_ngrid_t*, int row, char** rows, int size);
int g_rowncasecmp(a_ngrid_t*, int row, char** rows, int size);

int g_rownset(a_ngrid_t*, int row, char** values, int size);
int g_colnset(a_ngrid_t*, int col, char** values, int size);

char* g_strdup(a_ngrid_t*);

// default sorting is ascending and case-sensitive ( == 0);
// OR with these if you wish otherwise

#define SORT_DESCENDING 0x1
#define SORT_CASE_INSENSITIVE 0x2

int g_sortrow(a_ngrid_t*, int row, int flags);
int g_sortcol(a_ngrid_t*, int col, int flags);

int g_rowswap(a_ngrid_t*, int row1, int row2);
int g_colswap(a_ngrid_t*, int col1, int col2);

//-- forking and merging functions

a_ngrid_t* g_dup_rect(a_ngrid_t*, int top, int left,
                                  int height, int width);
a_ngrid_t* g_dup(a_ngrid_t*);
int        g_merge(a_ngrid_t* dst, a_ngrid_t* src);

int g_copyrow_add(a_ngrid_t* dst, a_ngrid_t* src, int row);

//-- and for you marco people out there

#define g_getceli g_getcellasint
#define g_setceli g_setcellasint
#define g_getcelt g_getcellastime
#define g_setcelt g_setcellastime
#define g_getcelf g_getcellasfloat
#define g_setcelf g_setcellasfloat


//-- the paths-abstraction layer

/*
 * A path is a grid that presents itself in this api as a tree.
 * Upon insertion of a value under a node, it gives you a new node, etc.
 * The file-format is again very readable (almost like output from tar -tvf)
 */

 /*
typedef struct nodestruct a_node_t;

struct nodestruct {
  a_grid_t* grid;
  a_node_t* parent;
  a_node_t* firstchild;
  a_node_t* nextbrother;
};

a_node_t* t_gettop(a_grid_t* grid);
void      t_free(a_node_t* top);

a_node_t* t_getparent(a_node_t* node);
a_node_t* t_getfirstchild(a_node_t* node);
a_node_t* t_getnextbrother(a_node_t* node);

a_node_t* t_addchild(a_node_t* node, char* value);

//-- experiment

typedef struct {
  char*** cells;
  int rowcount;
  int colcount;
  int rowoff;
  int coloff;
  int rowend;
  int colend;
} a_dbase_t;

a_dbase_t* d_create();
void       d_delete(a_dbase_t*);

int d_addtable(a_dbase_t*, char*);
int d_removetable(a_dbase_t*, char*);

int d_addfield(a_dbase_t*, char*);
int d_removefield(a_dbase_t*, char*);

int d_addrecord(a_dbase_t*);
int d_removerecord(a_dbase_t*, int);

int d_settable(a_dbase_t*, char*);
int d_nexttable(a_dbase_t*, char*);

char* d_gettablename(a_dbase_t*);
void  d_settablename(a_dbase_t*, char*);

char* d_getcell(a_dbase_t*, int record, char* field);
void  d_setcell(a_dbase_t*, int record, char* field, char* value);

*/

#ifdef __cplusplus
}
#endif

#define __A_GRID_H__
#endif
