/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: conf.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "conf.h"

char* cP_skip_whitespace(char* s) {
  while (*s && strchr("\n\r\t ", *s)) ++s;
  return s;
}

conf_node_t* cP_newnode(conf_t* conf, int parentindex) {
  conf_node_t* result;
  if (conf->count && !(conf->list[conf->count-1].name)) {
    conf->list[conf->count-1].parentindex = parentindex;
    return &(conf->list[conf->count-1]);
  }
  conf->list = (conf_node_t*)realloc(conf->list,
               sizeof(conf_node_t) * (conf->count+1));
  result = &(conf->list[conf->count++]);
  memset(result, 0, sizeof(*result));
  result->parentindex = parentindex;
  return result;
}

conf_node_t* cP_getnode(conf_t* conf, int index) {
  if (index >= 0 && index < conf->count) {
    return (&(conf->list[index]));
  }
  return NULL;
}

char* strndup(char* mem, int size, int dupifempty) {
  char* result;
  char swap = mem[size];
  mem[size] = 0;
  if (!strlen(mem) && !dupifempty) {
    result = NULL;
  } else {
    result = strdup(mem);
  }
  mem[size] = swap;
  return result;
}

char* cP_eos(char* string, char token) {
  do { ++string; } while (*string && *string != token);
  return string;
}

char* cP_eol(char* string) {
  while (*string) {
    if (*string == '\r') {
      if (*(string+1) == '\n') {
        return string+1;
      } else {
        return string;
      }
    } else if (*string == '\n') {
      return string;
    }
    ++string;
  }
  return string;
}

char* cP_eoi(char* string) {
  while (*string) {
    if (*string == '\\') {
      ++string;
    } else if (strchr(" \n\r\t{}=\"\'#", *string)) {
      return string;
    }
    ++string;
  }
  return string;
}

#define STAGE_NAME 0
#define STAGE_VALUE 1

int cP_parse(conf_t* conf, char* string) {
  char* s = cP_skip_whitespace(string), *b = s;
  if (*s) {
    conf_node_t* currentnode = cP_newnode(conf, -1);
    int stage = STAGE_NAME, parentindex = -1;
    while (*s) {
      switch (*s) {
      case '=':
        stage = STAGE_VALUE;
        s = cP_skip_whitespace(s+1);
        b = s;
        break;
      case '\"':
      case '\'':
        b = cP_eos(s, *s);
        switch (stage) {
        case STAGE_NAME:
          currentnode = cP_newnode(conf, parentindex);
          currentnode->name = strndup(s+1, b-(s+1), 1);
          break;
        case STAGE_VALUE:
          currentnode->value = strndup(s+1, b-(s+1), 1);
          stage = STAGE_NAME;
          break;
        }
        s = cP_skip_whitespace(b+1);
        b = s;
        break;
      case '{':
        parentindex = currentnode - conf->list;
        stage = STAGE_NAME;
        s = cP_skip_whitespace(s+1);
        b = s;
        break;
      case '}':
        currentnode = cP_getnode(conf, currentnode->parentindex);
        if (!currentnode) return -1;
        parentindex = currentnode->parentindex;
        stage = STAGE_NAME;
        s = cP_skip_whitespace(s+1);
        b = s;
        break;
      case '#':
        b = cP_eol(s);
        s = cP_skip_whitespace(b+1);
        b = s;
        break;
      default:
        b = cP_eoi(s);
        switch (stage) {
        case STAGE_NAME:
          currentnode = cP_newnode(conf, parentindex);
          currentnode->name = strndup(s, b-s, 0);
          break;
        case STAGE_VALUE:
          currentnode->value = strndup(s, b-s, 0);
          stage = STAGE_NAME;
          break;
        }
        s = cP_skip_whitespace(b);
        b = s;
        break;
      }
    }
  }
  return 0;
}

conf_t* c_create() {
  conf_t* result = (conf_t*)malloc(sizeof(conf_t));
  memset(result, 0, sizeof(*result));
  result->fd = -1;
  return result;
}

char* cP_slurp(int fd) {
  int off = 0, red;
  char* result = (char*)malloc(1024);
  if (!result) return NULL;
  while ((red = read(fd, result+off, 1023)) > 0) {
    off += red;
    result = (char*)realloc(result, off+1024);
    if (!result) return NULL;
  }
  result[off] = 0;
  return result;
}

conf_t* c_open(char* path) {
  if (path) {
    conf_t* result = c_create();
    if (result) {
      result->fd = open(path, O_RDWR|O_CREAT, 0664);
      if (result->fd >= 0) {
        char* buffer = cP_slurp(result->fd);
        if (!buffer) {
          close(result->fd);
          free(result); 
          return NULL;
        }
        if (strlen(buffer) && cP_parse(result, buffer)) {
          close(result->fd);
          free(buffer);
          free(result);
          return NULL;
        }
        free(buffer);
      }
    }
    return result;
  }
  return NULL;
}

conf_t* c_swallow(char* path) {
  conf_t* result = c_open(path);
  if (result) {
    close(result->fd);
    result->fd = -1;
  }
  return result;
}

conf_t* c_parse(char* string) {
  if (string) {
    conf_t* result = c_create();
    if (cP_parse(result, string)) {
      free(result);
      return NULL;
    }
    return result;
  }
  return NULL;
}

void c_close(conf_t* c) {
  if (c) {
    int i; for (i=0; i<c->count; i++) {
      if (c->list[i].name) free(c->list[i].name);
      if (c->list[i].value) free(c->list[i].value);
    }
    free(c->list);
    c->count = 0;
    c->list = NULL;
    free(c);
  }
}

void c_closefd(conf_t* c) {
  if (c) {
    if (c->fd >= 0) {
      close(c->fd);
    }
  }
}

int* cP_dupchildlist(conf_t* conf, int parentindex, int* len) {
  int i, cnt = 32, * result = (int*)malloc(sizeof(int) * 32);
  *len = 0;
  for (i=0; i<conf->count; i++) {
    if (parentindex < 0 ||
        parentindex >= conf->count) {
      if (conf->list[i].parentindex < 0 ||
          conf->list[i].parentindex >= conf->count) {
        if (*len >= cnt) {
          result = (int*)realloc(result, sizeof(int) * (cnt += 32));
          if (!result) {
            *len = 0;
            return NULL;
          }
        }
        result[(*len)++] = i;
      }
    } else {
      if (conf->list[i].parentindex == parentindex) {
        if (*len >= cnt) {
          result = (int*)realloc(result, sizeof(int) * (cnt += 32));
          if (!result) {
            *len = 0;
            return NULL;
          }
        }
        result[(*len)++] = i;
      }
    }
  }
  return result;
}

void cP_escape(char* out, char* in) {
  if (!strlen(in)) {
    strcpy(out, "\"\"");
    return;
  } else {
    int quoted = 0;
    char* i = in, * o = out;
    while (*i) {
      if (strchr("\'\"", *i)) {
        *o++ = '\\';
        quoted = 1;
      } else if (strchr("={}@?. \n\r\t", *i)) {
        quoted = 1;
      }
      *o++ = *i++;
    }
    *o++ = 0;
    if (quoted) {
      memmove(out+1, out, (o-out));
      *out = '\"';
      strcat(out, "\"");
    }
  }
}

char* cP_catitem(conf_t* conf, char* string, int* off, int* len,
                 char* cat, int spc) {
  // maximum possible escaped, quoted item length
  int itemlen = spc + (strlen(cat) * 2) + 3;
  while (*off + itemlen >= *len) {
    string = (char*)realloc(string, ((*len) += 1024));
  }
  memset(string + *off, ' ', spc);
  cP_escape(string + *off + spc, cat);
  *off += strlen(string + *off);
  return string;
}

char* cP_catctrl(conf_t* conf, char* string, int* off, int* len,
                 char* cat, int spc) {
  int itemlen = spc + strlen(cat) + 1;
  while (*off + itemlen >= *len) {
    string = (char*)realloc(string, ((*len) += 1024));
  }
  memset(string + *off, ' ', spc);
  strcpy(string + *off + spc, cat);
  *off += strlen(string + *off);
  return string;
}

char* cP_strdup(conf_t* conf, char* string, int* off, int* len, int idx,
                int lvl, int tight) {
  conf_node_t* node = &(conf->list[idx]);
  int l, * c = cP_dupchildlist(conf, idx, &l);
  if (node->name) {
    string = cP_catitem(conf, string, off, len, node->name, lvl);
    if (node->value) {
      string = cP_catctrl(conf, string, off, len, "=", (tight ? 0 : 1));
      string = cP_catitem(conf, string, off, len, node->value, (tight ? 0 : 1));
    }
    if (l > 0) {
      int i;
      string = cP_catctrl(conf, string, off, len, "{\n", (tight ? 0 : 1));
      for (i=0; i<l; i++) {
        string = cP_strdup(conf, string, off, len, c[i], lvl+1, tight);
      }
      string = cP_catctrl(conf, string, off, len, "}\n", (tight ? 0 : lvl));
    } else {
      if (tight) {
        string = cP_catctrl(conf, string, off, len, " ", 0);
      } else {
        string = cP_catctrl(conf, string, off, len, "\n", 0);
      }
    }
  }
  free(c);
  return string;
}

void cP_dup(conf_t* orig, int* list, int len, conf_t* copy, int ncopy) {
  int i;
  for (i=0; i<len; i++) {
    int* childlist,
         childlistlen,
         nchild = c_add_child(copy, ncopy,
                  orig->list[list[i]].name, orig->list[list[i]].value);
    childlist = cP_dupchildlist(orig, list[i], &childlistlen);
    cP_dup(orig, childlist, childlistlen, copy, nchild);
    free(childlist);
  }
}

conf_t* c_copy(conf_t* original, int node) {
  if (original) {
    conf_t* result = c_create();
    if (result) {
      int len, * list = cP_dupchildlist(original, node, &len);
      cP_dup(original, list, len, result, -1);
      free(list);
    }
    return result;
  }
  return NULL;
}

char* c_strdup(conf_t* conf, int tight) {
  if (conf) {
    char* string = (char*)malloc(1024);
    int off = 0, len = 1024;
    int i, l, * c = cP_dupchildlist(conf, -1, &l);
    *string = 0;
    for (i=0; i<l; i++) {
      string = cP_strdup(conf, string, &off, &len, c[i], 0, tight);
    }
    return string;
  }
  return NULL;
}

int cP_write(conf_t* conf) {
  if (conf) {
    char* string = c_strdup(conf, 0);
    write(conf->fd, string, strlen(string)); //!
    free(string);
    return 0;
  }
  return -1;
}

int c_save(conf_t* conf) {
  if (conf && conf->fd >= 0) {
    ftruncate(conf->fd, 0);
    lseek(conf->fd, 0, SEEK_SET);
    return cP_write(conf);
  }
  return -1;
}

char* c_name(conf_t* conf, int node) {
  return ((conf && node >= 0 && node < conf->count) ? conf->list[node].name
                                                    : NULL);
}

char* c_value(conf_t* conf, int node) {
  return ((conf && node >= 0 && node < conf->count) ? conf->list[node].value
                                                    : NULL);
}

int c_value_int(conf_t* conf, int node) {
  if (conf &&
      node >= 0 && node < conf->count &&
      conf->list[node].value) {
    return atoi(conf->list[node].value);
  }
  return 0;
}

int c_first_child(conf_t* conf, int idx) {
  if (conf) {
    if (idx >= 0 && idx < conf->count) {
      int i;
      for (i=0; i<conf->count; i++) {
        if (conf->list[i].parentindex == idx) return i;
      }
    } else {
      int i; for (i=0; i<conf->count; i++) {
        if (conf->list[i].parentindex < 0 ||
            conf->list[i].parentindex >= conf->count) return i;
      }
    }
  }
  return -1;
}

int c_next_brother(conf_t* conf, int idx) {
  if (conf && idx >= 0 && idx < conf->count) {
    int i;
    for (i=idx+1; i<conf->count; i++) {
      if (conf->list[idx].parentindex < 0 ||
          conf->list[idx].parentindex >= conf->count) {
        if (conf->list[i].parentindex < 0 ||
            conf->list[i].parentindex >= conf->count) {
          return i;
        }
      } else {
        if (conf->list[i].parentindex == conf->list[idx].parentindex) {
          return i;
        }
      }
    }
  }
  return -1;
}

int c_first_child_named(conf_t* conf, int idx, char* name) {
  if (conf && name) {
    int result;
    if ((result = c_first_child(conf, idx)) != -1) {
      if (!strcmp(conf->list[result].name, name)) return result;
      while ((result = c_next_brother(conf, result)) != -1) {
        if (!strcmp(conf->list[result].name, name)) return result;
      }
    }
  }
  return -1;
}

int c_next_brother_named(conf_t* conf, int idx, char* name) {
  if (conf && idx >= 0 && idx < conf->count && name) {
    while ((idx = c_next_brother(conf, idx)) != -1) {
      if (!strcmp(conf->list[idx].name, name)) return idx;
    }
  }
  return -1;
}

int c_first_child_casenamed(conf_t* conf, int idx, char* name) {
  if (conf && name) {
    int result;
    if ((result = c_first_child(conf, idx)) != -1) {
      if (!strcasecmp(conf->list[result].name, name)) return result;
      while ((result = c_next_brother(conf, result)) != -1) {
        if (!strcasecmp(conf->list[result].name, name)) return result;
      }
    }
  }
  return -1;
}

int c_next_brother_casenamed(conf_t* conf, int idx, char* name) {
  if (conf && idx >= 0 && idx < conf->count && name) {
    while ((idx = c_next_brother(conf, idx)) != -1) {
      if (!strcasecmp(conf->list[idx].name, name)) return idx;
    }
  }
  return -1;
}

char* c_first_child_value_named(conf_t* conf, int node, char* name) {
  if (conf) {
    int idx = c_first_child_named(conf, node, name);
    return c_value(conf, idx);
  }
  return NULL;
}

char* c_next_brother_value_named(conf_t* conf, int node, char* name) {
  if (conf) {
    int idx = c_next_brother_named(conf, node, name);
    return c_value(conf, idx);
  }
  return NULL;
}

int c_add_child(conf_t* conf, int node, char* name, char* value) {
  if (conf && name) {
    conf_node_t* child = cP_newnode(conf, node);
    if (!child) return -1;
    child->name = strdup(name);
    if (value) {
      child->value = strdup(value);
    }
    return child - conf->list;
  }
  return -1;
}

int c_add_brother(conf_t* conf, int node, char* name, char* value) {
  if (conf && name && node >= 0 && node < conf->count) {
    return c_add_child(conf, conf->list[node].parentindex, name, value);
  }
  return -1;
}

int c_set(conf_t* conf, int node, char* value) {
  if (conf && node >= 0 && node < conf->count) {
    if (conf->list[node].value) free(conf->list[node].value);
    conf->list[node].value = (value ? strdup(value) : NULL);
    return 0;
  }
  return -1;
}

int c_add_child_int(conf_t* conf, int node, char* name, int value) {
  char string[64];
  snprintf(string, 64, "%d", value);
  return c_add_child(conf, node, name, string);
}

int c_add_brother_int(conf_t* conf, int node, char* name, int value) {
  char string[64];
  snprintf(string, 64, "%d", value);
  return c_add_brother(conf, node, name, string);
}

char* c_name_strncpy(conf_t* conf, int node, char* buffer, int size) {
  if (conf && node >= 0 && node < conf->count &&
      conf->list[node].name && buffer && size >= 0) {
    strncpy(buffer, conf->list[node].name, size);
  }
  return buffer;
}

char* c_value_strncpy(conf_t* conf, int node, char* buffer, int size) {
  if (conf && node >= 0 && node < conf->count &&
      conf->list[node].value && buffer && size >= 0) {
    strncpy(buffer, conf->list[node].value, size);
  }
  return buffer;
}

void cP_next_node(char* string, int* begin, int* end) {
  char* s = string + (*begin);
  while (*s) {
    if (*s == '\\') {
      ++s;
    } else if (*s == '.') {
      *end = s - string;
      return;
    }
    ++s;
  }
  *end = s - string;
}

/*
* a node may typically be:
* (<name>|'@') [ '[' (<index>|'*') ']' ]
* ['?'<value> [ '[' (<index>|'*') ']' ] ]
*/

struct nodesplit {
  char* name;
  char* nameindex;
  char* value;
  char* valueindex;
};

int cP_node_split(char* string, struct nodesplit* nsplit) {
  char* s = cP_skip_whitespace(string);
  memset(nsplit, 0, sizeof(struct nodesplit));
  if (!(*s)) return -1; // no name
  nsplit->name = s;
  if (*s == '@') {
    ++s;
  } else {
    while (*s) {
      if (*s == '\\') {
        ++s;
      } else if (strchr("?[", *s)) {
        break;
      }
      ++s;
    }
  }
  s = cP_skip_whitespace(s);
  if (!(*s)) {
    return 0; // only a name
  }
  if (*s == '[') {
    *s = 0;
    s = cP_skip_whitespace(s+1);
    nsplit->nameindex = s;
    while (*s) {
      if (*s == '\\') {
        ++s;
      } else if (*s == ']') {
        break;
      }
      ++s;
    }
    if (!(*s)) {
      return -1; // no closing ']'
    } else {
      *s = 0;
      s = cP_skip_whitespace(s+1);
      if (!(*s)) {
        return 0; // name + nameindex
      }
    }
  }
  if (*s == '?') {
    *s = 0;
    s = cP_skip_whitespace(s+1);
    if (!(*s)) {
      return -1; // no value
    }
    nsplit->value = s;
    while (*s) {
      if (*s == '\\') {
        ++s;
      } else if (*s == '[') {
        break;
      }
      ++s;
    }
    if (!(*s)) {
      return 0;  // value
    }
    if (*s == '[') {
      *s = 0;
      s = cP_skip_whitespace(s+1);
      nsplit->valueindex = s;
      while (*s) {
        if (*s == '\\') {
          ++s;
        } else if (*s == ']') {
          break;
        }
        ++s;
      }
      if (!(*s)) {
        return -1; // no closing ']'
      } else {
        *s = 0;
        s = cP_skip_whitespace(s+1);
        if (!(*s)) {
          return 0; // value + valueindex
        } else {
          return -1; // trailing chars
        }
      }
    }
  }
  return -1; // trailing chars, impossible situations and the like
}

#define NAME  0
#define VALUE 1
#define COUNT 2

int cP_evaluate_node(conf_t* conf, char* string,
                     int pidx, int* idx, int* part) {
  struct nodesplit nsplit;
  if (*part == COUNT) return -1;
  if (cP_node_split(string, &nsplit) != 0) return -1;
  if (nsplit.name && *(nsplit.name) == '@') {
    if (!nsplit.nameindex) return -1;
    if (*(nsplit.nameindex) == '*') { // count of any name
      int i, cnt = 0;
      for (i=0; i<conf->count; i++) {
        if (conf->list[i].parentindex == pidx) {
          ++cnt;
        }
      }
      *part = COUNT;
      return cnt;
    } else { // instance of indexed name
      int index = atoi(nsplit.nameindex);
      int cnt = 0, i;
      for (i=0; i<conf->count; i++) {
        if (conf->list[i].parentindex == pidx) {
          if (cnt++ == index) {
            *idx = i;
            *part = NAME;
            return 0;
          }
        }
      }
      return -1;
    }
  } else {
    int nameindex = -1, valueindex = -1;
    if (nsplit.nameindex) {
      if (*(nsplit.nameindex) != '*') {
        nameindex = atoi(nsplit.nameindex);
      }
    } else {
      nameindex = 0;
    }
    if (nsplit.valueindex) {
      if (*(nsplit.valueindex) != '*') {
        valueindex = atoi(nsplit.valueindex);
      }
    } else {
      valueindex = 0;
    }

    if (nameindex == -1) {
      int cnt = 0, i;
      if (nsplit.value) {
        if (nsplit.valueindex) { // "foo[*]?bar[2]" "foo[*]?bar[*]" ??!!??
          return -1;
        }
        // "foo[*]?bar"
        for (i=0; i<conf->count; i++) {
          if (!strcmp(conf->list[i].name, nsplit.name) &&
              conf->list[i].value &&
              !strcmp(conf->list[i].value, nsplit.value)) {
            ++cnt;
          }
        }
        *part = COUNT;
        return cnt;
      }
      // "foo[*]"
      for (i=0; i<conf->count; i++) {
        if (!strcmp(conf->list[i].name, nsplit.name)) {
          ++cnt;
        }
      }
      *part = COUNT;
      return cnt;
    } else {
      int cnt = 0, i;
      if (nsplit.value) {
        if (valueindex == -1) { // "foo?bar[*]"
          for (i=0; i<conf->count; i++) {
            if (!strcmp(conf->list[i].name, nsplit.name) &&
                conf->list[i].value &&
                !strcmp(conf->list[i].value, nsplit.value)) {
              ++cnt;
            }
          }
          *part = COUNT;
          return cnt;
        }
        // "foo?bar[2]"
        for (i=0; i<conf->count; i++) {
          if (!strcmp(conf->list[i].name, nsplit.name) &&
              conf->list[i].value &&
              !strcmp(conf->list[i].value, nsplit.value)) {
            if (valueindex == cnt++) {
              *part = VALUE; 
              *idx = i;
              return 0;
            }
          }
        }
      }
      // "foo"
      for (i=0; i<conf->count; i++) {
        if (!strcmp(conf->list[i].name, nsplit.name)) {
          *part = VALUE;
          *idx = i;
          return 0;
        }
      }
      return -1;
    }
  }
  return -1;
}

int cP_query(conf_t* conf, char* string, int* node, int* part) {
  int result, begin = 0, end, parentindex = -1;
  for (;;) {
    char swap;
    cP_next_node(string, &begin, &end);
    swap = string[end];
    string[end] = 0;
    if ((result = cP_evaluate_node(conf, string + begin, parentindex,
                                   node, part)) != 0) {
      string[end] = swap;
      return result;
    }
    parentindex = *node;
    if (swap == 0) break;
    string[end] = swap;
    begin = end+1;
  }
  return 0;
}

char* c_query(conf_t* conf, char* query) {
  if (conf && query) {
    char* string = strdup(query);
    static char buffer[32];
    int node, part = -1;
    int cnt = cP_query(conf, string, &node, &part);
    switch (part) {
    case NAME:
      return ((cnt == 0) ? conf->list[node].name : NULL);
    case VALUE:
      return ((cnt == 0) ? conf->list[node].value : NULL);
    case COUNT:
      snprintf(buffer, 32, "%d", cnt);
      return buffer;
    }
  }
  return NULL;
}

#include <stdarg.h>

char* c_qsprintf(conf_t* conf, char* fmt, ...) {
  if (conf && fmt) {
    char* result, buffer[4096];
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buffer, 4096, fmt, ap);
    buffer[4095] = 0;
    result = c_query(conf, buffer);
    va_end(ap);
    return result;
  }
  return NULL;
}


#ifdef __CONFIG_C_MAIN__

int main(int argc, char** argv) {
  int i;
  char* cf = strdup(
    "server=\"192.168.1.202\" {\n"
    "  port=9032\n"
    "  disallow {\n"
    "    address=\"192.168.1.19\"\n"
    "    another andanother andyetanother\n"
    "  }\n"
    "}\n"
    "oi\n"
    "server=\"192.168.1.202\" {\n"
    "  port=9032\n"
    "  disallow {\n"
    "    address=\"192.168.1.19\"\n"
    "    address=\"192.168.1.250\"\n"
    "    another=\"very {} difficult\"\n"
    "  }\n"
    "}\n"
    );
  conf_t* c = c_parse(cf);
//  printf("\n");
//  for (i=0; i<c->count; i++) {
//    printf("'%s' '%s' %d\n", c->list[i].name, c->list[i].value, c->list[i].parentindex);
//  }
//  printf(c_strdup(c, 0));
  printf(c_query(c, "server[*]"));
  return 0;
}

#endif
