/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * $Id: httpc.h,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#ifndef _HTTPC_H_
#define _HTTPC_H_

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <sys/time.h>
#include <pthread.h>

#include <regex.h>

#include "httpc_parse.h"
#include "conf.h"
#include "a_grid.h"

#include <sys/types.h>
#include <limits.h>
#ifdef _OS_LINUX
#include <db1/db.h>
#else
#include <db.h>
#endif // _OS_LINUX

#define VERSION_MAJOR 0
#define VERSION_MINOR 5

// please set these values sensibly suited to your machine's needs and caps !

#define MAXFILESIZE    (64*1024)
#define MAXHEADERSIZE  (64*1024)

// maxthreads is a default if not given (or zero given) 'maxthreads'
// in the config.  The load balancing watcher thread uses this given.

#define HTTPC_MAXTHREADS 20
#define HTTPC_MINTHREADS 4

typedef enum {
  IDLE,             // set initially by httpc-thread and then by handler-thread
  BUSY,             // set by handler-thread
  DYING             // set by httpc-thread
} httpc_handler_state;

typedef struct _httpc_box httpc_box_t;
typedef struct _httpc_server httpc_server_t;
typedef struct _session session_t;
typedef struct _client client_t;
typedef struct _httpc httpc_t;

typedef struct {
  httpc_t*            httpc;
  httpc_handler_state state;
  client_t*           client;    // when I take on a client
  time_t              timestamp; // when I take on a client
} httpc_handler_t;

typedef struct {
  httpc_handler_t** list;
  int               count;
  pthread_mutex_t   mutex;
  pthread_mutex_t   condmutex;
  pthread_cond_t    condition;
} httpc_handlerlist_t;

typedef enum {
  CAPABLE,            // netscape
  JAVASCRIPT_CAPABLE, // msie
  INCAPABLE           // lynx
} browser_t;

#define BOXNAMESIZE 16

struct _session {
  httpc_box_t*      box;
  time_t            timestamp;
  int               clientrefcount;
  struct in_addr    address;
  browser_t         browser;
  conf_t*           var;
  char*             boxname; // pointer in var
  char*             cookie;  // pointer in var
  char*             user;    // pointer in var
};

typedef struct {
  session_t**     list;
  int             count;
  pthread_mutex_t mutex;
} sessionlist_t;

struct passwdfile {
  a_ngrid_t* table;
};

union authtypes {
  struct passwdfile pwdf;
};

#define HTTPC_AUTH_TABLE 1

struct auth {
  int use_auth;
  int type;
  union authtypes types;
  int allow_anonymous;
};

typedef enum {
  SILENT = 0,
  SELECTABLE,
  SELECTED,
  HANDLED
} iostate_t;

struct _client {
  int                fd;      // connection fd
  struct in_addr     address; // client IP address
#ifdef _DEBUG
  int                port;    // in case of debug, the client port is handy
#endif
  iostate_t          readstate;
  iostate_t          writestate;
  int                sucking;
  httpc_server_t*    server;  // the server that accepted the connection
  httpc_handler_t*   handler; // my handler resets this one to NULL when done
  session_t*         session; // a request's cookie will resolve a session
  struct httprequest request;
};

typedef struct {
  client_t**      list;
  int             count;
  pthread_mutex_t mutex;
} clientlist_t;

typedef struct {
  struct in_addr* list;
  int             count;
} rejectlist_t;

struct _httpc_box {
  regex_t         preg;    // to match '/joins', the default matches ""
  char*           boxname; // if set, matches to PREG will enforce this name
  int             maxlines;
  time_t          time;
  char*           head;        // template for html head
  char*           bodyprefix;  // template for html body prefix
  char*           bodypostfix; // template for html body postfix
  char*           line;        // template for chatline
  char*           personal;    // template for personal chatline
  char*           message;     // template for message
  conf_t*         var;         // preset context
};

typedef struct {
  httpc_box_t* list;
  int          count;
} boxlist_t;

struct _httpc_server {
  int                 logfd;
  char*               name;
  httpc_t*            httpc;
  struct sockaddr     ipaddress;
  unsigned short      port;
  int                 sockfd;    // fd to listening socket
  clientlist_t        clients;
  int                 serverrootnode;
  rejectlist_t        rejects;
  pthread_mutex_t     boxmutex;
  DB*                 db;
  struct auth         auth;
  conf_t*             files;
  short               musthavecookie;   // defaults to 0
  short               useforeigncookie; // defaults to 0
  boxlist_t           boxes;
};

typedef struct {
  httpc_server_t** list;
  int              count;
} httpc_serverlist_t;

typedef enum {
  APPLICATION_START = 1,
  APPLICATION_RUNNING,
  APPLICATION_END
} httpc_state_t;

struct watch {
  pthread_t thread;
  int       visits;
  int       visits_1;
  int       recurrence;
  int       maxthreads;
  int       minthreads;
  int       sleeptime;
};

struct _httpc {
  httpc_state_t       state;
  conf_t*             config;
  int                 pipe[2];
  httpc_serverlist_t  servers;
  httpc_handlerlist_t handlers;
  sessionlist_t       sessions;
  int                 adminsockfd;
  struct watch        watch;
  int                 waitcount; // number of clients waiting to be handled
};

#define HTTPC_SIGNEW  0
#define HTTPC_SIGSTOP 1

httpc_t* httpc_init(char* path);
int      httpc_start(httpc_t* httpc);
void     httpc_stop(httpc_t* httpc);

client_t* httpc_add_client(httpc_server_t* s, int fd, struct in_addr a);
void httpc_close_client_safe(client_t* c);
void httpc_close_client(client_t* c);
void httpc_remove_client_safe(httpc_server_t* s, client_t* client);
void httpc_remove_client(httpc_server_t* s, client_t* client);

int httpc_spawn_handler(httpc_t* t);
int httpc_examine_handlers(httpc_t* t, int treshold);

#endif
