/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_admin.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "httpc_admin.h"
#include "httpc_session.h"
#include "conf.h"

int httpc_admin_init(httpc_t* t) {
  int n;
  if ((n = c_first_child_named(t->config, -1, "admin")) != -1) {
    char* path = c_value(t->config, n);
    if (path) {
      unlink(path);
      t->adminsockfd = socket(AF_UNIX, SOCK_STREAM, 0);
      if (t->adminsockfd >= 0) {
        struct sockaddr_un unixserver;
        unixserver.sun_family = AF_UNIX;
        strcpy(unixserver.sun_path, path);
        if (bind(t->adminsockfd,
            (struct sockaddr*)&unixserver,
            SUN_LEN(&unixserver))) {
          perror("httpc_admin_init:bind");
          return -1;
        }
        if (listen(t->adminsockfd, 1)) {
          perror("httpc_admin_init:bind");
          return -1;
        }
        fprintf(stderr, "Admin init successfully using socket '%s'.\n", path);
        return 0;
      }
    }
  }
  return -1;
}

int hasended(char* string, int len) {
  while (len >= 0 && string[len] <= ' ') {
    --len;
  }
  if (len > 2 &&
      string[len] == '=' &&
      (string[len-1] == '\n' || string[len-1] == '\r')) {
    string[len] = ' ';
    return 1;
  }
  return 0;
}

// there are two ways to indicate an end of streamdata here;
// a zero char or an '=' on an empty line.

char* suckdup(int fd) {
  int off = 0, n;
  char* result = (char*)malloc(1025);
  while ((n = read(fd, result + off, 1024)) >= 0) {
    off += n;
    if (result[off] == 0) return result;
    result[off] = 0;
    if (hasended(result, off)) return result;
    result = (char*)realloc(result, (off + 1025));
  }
  free(result);
  return NULL;
}

int httpcP_admin_createsession(httpc_t* t, conf_t* request) {
  char* cookie = c_first_child_value_named(request, -1, "cookie");
  if (cookie) {
    int n = c_first_child(request, -1);
    session_t* session = httpc_add_session(t);
    httpc_set_cookie(session, cookie);
    while (n != -1) {
      char* name = c_name(request, n);
      char* value = c_value(request, n);
      if (name && value) {
        if (!strcmp(name, "user")) {
          httpc_set_user(session, value);
        } else if (strcmp(name, "cookie")) {
          c_add_child(session->var, -1, name, value);
        }
      }
      n = c_next_brother(request, n);
    }
#ifdef _DEBUG
    fprintf(stderr, "Admin socket created session '%s'\n", cookie);
#endif
    return 0;
  }
  return -1;
}

int httpcP_handle_admin_request(httpc_t* t, conf_t* request) {
  int n;
  if ((n = c_first_child_named(request, -1, "action")) != -1) {
    char* value = c_value(request, n);
    if (value) {
      if (!strcmp(value, "createsession")) {
        return httpcP_admin_createsession(t, request);
      } else if (!strcmp(value, "killsession")) {
      } else if (!strcmp(value, "endsession")) {
      } else if (!strcmp(value, "chat")) {
      } else if (!strcmp(value, "message")) {
      }
    }
  }
  if ((n = c_first_child_named(request, -1, "request")) != -1) {
    char* value = c_value(request, n);
    if (value) {
      if (!strcmp(value, "sessionlist")) {
      } else if (!strcmp(value, "userlist")) {
      }
    }
  }
  return -1;
}

int httpc_admin_accept(httpc_t* t) {
  if (t->adminsockfd >= 0) {
    struct sockaddr_un client;
    int fd, cl = 0;
    memset(&client, 0, sizeof(client));
    fd = accept(t->adminsockfd, (struct sockaddr*)&client, &cl);
#ifdef _DEBUG
    fprintf(stderr, "Admin socket addressed.\n");
#endif
    if (fd >= 0) {
      char* string = suckdup(fd);
#ifdef _DEBUG
      fprintf(stderr, "Admin socket sends '%s'.\n", string);
#endif
      if (string) {
        conf_t* conf = c_parse(string);
        if (conf) {
          int result = httpcP_handle_admin_request(t, conf);
          c_close(conf);
          free(string);
          close(fd);
          return result;
        }
        free(string);
      }
      close(fd);
    }
  }
  return -1;
}
