/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_box.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
//#ifdef OS_LINUX
#define _GNU_SOURCE
//#endif

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "httpc_box.h"
#include "httpc_util.h"
#include "httpc_format.h"
#include "httpc_session.h"
#include "httpc_db.h"
#include "a_grid.h"

void httpcP_add_variables(httpc_box_t* box, conf_t* config, int node) {
  int nconfig = c_first_child(config, node);
  while (nconfig != -1) {
    char* name = c_name(config, nconfig);
    char* value = c_value(config, nconfig);
    int nvar = c_first_child_named(box->var, -1, name);
    if (nvar != -1) {
      c_set(box->var, nvar, value);
    } else {
      c_add_child(box->var, -1, name, value);
    }
    nconfig = c_next_brother(config, nconfig);
  }
}

void httpcP_add_files(httpc_server_t* server, conf_t* config, int node) {
  int nconfig = c_first_child(config, node);
  while (nconfig != -1) {
    char* name = c_name(config, nconfig);
    char* value = c_value(config, nconfig);
    if (value) {
      char* content = fstrdup(value, MAXFILESIZE);
      if (content) {
        int nfile = c_first_child_named(server->files, -1, name);
        if (nfile != -1) {
          c_set(server->files, nfile, content);
        } else {
          c_add_child(server->files, -1, name, content);
        }
        free(content);
      }
    }
    nconfig = c_next_brother(config, nconfig);
  }
}

void httpcP_basicauth_table(httpc_server_t* server, char* path) {
  if ((server->auth.types.pwdf.table = g_open(path))) {
    server->auth.use_auth = 1;
    server->auth.type = HTTPC_AUTH_TABLE;
    fprintf(stderr, "Server '%s' has basic authentication from '%s'.\n",
                    server->name, path);
  }
}

void httpcP_fill_server_config(httpc_server_t* server,
                               conf_t* config, int node) {
  int n = c_first_child(config, node);
  while (n != -1) {
    char* name = c_name(config, n), * value = c_value(config, n);
    if (name) {
      if (value) {
        if (!strcmp(name, "log")) {
          fprintf(stderr, "Server logged in '%s'.\n", value);
          server->logfd = open(value, O_WRONLY|O_CREAT|O_APPEND, 0644);
          if (server->logfd >= 0) {
            char line[] = "\nNew session.\n";
            write(server->logfd, line, strlen(line));
          }
        } else if (!strcmp(name, "name")) {
          server->name = value;
        } else if (!strcmp(name, "authentication")) {
          if (c_first_child(config, n) != -1) {
            //.. options may include 'type', 'anonymous'
          } else {
            httpcP_basicauth_table(server, value);
          }
        } else if (!strcmp(name, "behaviour")) {
          if (!strcmp(value, "use_foreign_cookies")) {
            server->useforeigncookie = 1;
          } else if (!strcmp(value, "must_have_cookie")) {
            server->musthavecookie = 1;
          }
        } else if (!strcmp(name, "state")) {
          httpc_db_init(server, value);
        }
      } else if (!strcmp(name, ".files")) {
        httpcP_add_files(server, config, n);
      }
    }
    n = c_next_brother(config, n);
  }
  if (!server->db) {
    httpc_db_init(server, NULL);
  }
}

void httpcP_fill_box_config(httpc_box_t* box, conf_t* config, int node) {
  int n;
  box->boxname     = NULL;
  box->maxlines    = 30;
  box->time        = time(NULL);
  box->head        = NULL;
  box->bodyprefix  = NULL;
  box->bodypostfix = NULL;
  box->line        = NULL;
  box->personal    = NULL;
  box->message     = NULL;
  box->var         = c_create();
  n = c_first_child(config, node);
  while (n != -1) {
    char* name = c_name(config, n), * value = c_value(config, n);
    if (name) {
      if (value) {
        if (!strcmp(name, "boxname")) {
          box->boxname = value;
        } else if (!strcmp(name, "head")) {
          box->head = fstrdup(value, MAXFILESIZE);
        } else if (!strcmp(name, "bodyprefix")) {
          box->bodyprefix = fstrdup(value, MAXFILESIZE);
        } else if (!strcmp(name, "bodypostfix")) {
          box->bodypostfix = fstrdup(value, MAXFILESIZE);
        } else if (!strcmp(name, "line")) {
          box->line = fstrdup(value, MAXFILESIZE);
        } else if (!strcmp(name, "personal")) {
          box->personal = fstrdup(value, MAXFILESIZE);
        } else if (!strcmp(name, "message")) {
          box->message = fstrdup(value, MAXFILESIZE);
        } else if (!strcmp(name, "maxlines")) {
          box->maxlines = atoi(value);
        }
      } else if (!strcmp(name, ".var")) {
        httpcP_add_variables(box, config, n);
      }
    }
    n = c_next_brother(config, n);
  }
  if (!box->line) {
    box->line = strdup("${user}:${line}<br>\n");
  }
  if (!box->personal) {
    box->personal = strdup("<b>${user}:${line}</b><br>\n");
  }
  if (!box->message) {
    box->message = strdup("<i>${user}:${line}</i><br>\n");
  }
}

void httpc_add_serverinfo(httpc_server_t* server, int node) {

  conf_t* config = server->httpc->config;
  int n;

  server->logfd       = -1;
  server->files       = c_create();
  httpcP_fill_server_config(server, config, node);

  server->boxes.list  = (httpc_box_t*)malloc(sizeof(httpc_box_t));
  server->boxes.count = 1;
  httpcP_fill_box_config(&(server->boxes.list[0]), config, node);

  n = c_first_child(config, node);
  while (n != -1) {
    char* name = c_name(config, n), * value = c_value(config, n);
    if (name && value && !strcmp(name, "box")) {
      regex_t preg;
      if (regcomp(&preg, value, 0) == 0) {
        httpc_box_t* box;
        server->boxes.list = (httpc_box_t*)realloc(server->boxes.list,
                             sizeof(httpc_box_t) * (server->boxes.count+1));
        box = &(server->boxes.list[server->boxes.count++]);
        httpcP_fill_box_config(box, config, n);
        memcpy(&(box->preg), &preg, sizeof(regex_t));
        if (box->boxname) {
          fprintf(stderr, "Added box '%s' as '%s'.\n", value, box->boxname);
        } else {
          fprintf(stderr, "Added box '%s'.\n", value);
        }
      } else {
        fprintf(stderr, "Could not compile regular expression '%s'.\n", value);
      }
    }
    n = c_next_brother(config, n);
  }
}

/*
* Adds a chatline with known recipient to this client's box's buffer.
*/

void httpc_add_line(client_t* c, char* line, char* recipient) {
  char* hashstring;
  conf_t* hash = c_copy(c->session->var, -1);
  c_add_child(hash, -1, "line", line);
  if (recipient) {
    c_add_child(hash, -1, "recipient", recipient);
  }
  hashstring = c_strdup(hash, ~0);
  httpc_db_insert(c, hashstring);
  free(hashstring);
  c_close(hash);
}

/*
* Adds a chat line to the client's box's buffer.
*/

void httpc_add_chat(client_t* c, char* line) {
  httpc_add_line(c, line, NULL);
}

/*
* Interprets a chatline that starts with '/join'
*/

void httpcP_interpret_join(client_t* c, char* relbox) {
  if (relbox) {
    int i;
    for (i=1; i<c->server->boxes.count; i++) {
      if (regexec(&(c->server->boxes.list[i].preg), relbox, 0, NULL, 0) == 0) {
        if (c->server->boxes.list[i].boxname) {
          httpc_set_boxname(c->session, c->server->boxes.list[i].boxname);
        } else {
          httpc_set_boxname(c->session, relbox);
        }
        c->session->box = &(c->server->boxes.list[i]);
        return;
      }
    }
    httpc_set_boxname(c->session, relbox);
    c->session->box = &(c->server->boxes.list[0]);
  } else {
    httpc_set_boxname(c->session, "");
    c->session->box = &(c->server->boxes.list[0]);
  }
}

/*
* Interprets a chatline that starts with '/msg'
*/

void httpcP_interpret_msg(client_t* c, char* recipient) {
  if (recipient) {
    char* line = skip_identifier(recipient);
    if (*line) {
      *line = 0;
      httpc_add_line(c, line+1, recipient);
    }
  }
}

/*
* Lines may start with a slash command, in which case the line must
* be _interpreted_ rather than fed to the lines table right away.
* Valid commands so far are:
* '/join' [S]+ [boxname]
* '/msg' [S]+ [username] [S]+ [line]
*/

void httpc_interpret_line(client_t* c, char* line) {
  if (*line != '/') {
    httpc_add_line(c, line, NULL);
    return;
  }
  if (!strncasecmp(line+1, "join", 4)) {
    httpcP_interpret_join(c, skip_whitespace(line+5));
  } else if (!strncasecmp(line+1, "msg", 3)) {
    httpcP_interpret_msg(c, skip_whitespace(line+4));
  } else {
    httpc_add_line(c, line, NULL);
  }
}

/*
* Returns non-zero if client info matches cookie or user.
*/

int httpcP_is_personal(client_t* client, char* cookie, char* user) {
  if (cookie && client->session->cookie) {
    if (!strcmp(cookie, client->session->cookie)) return 1;
  }
  if (user && client->session->user) {
    if (!strcmp(user, client->session->user)) return 1;
  }
  return 0;
}

/*
* Returns a malloced string containing the chatlines from this
* client's box's buffer.  It is also slightly relative to this client
* in that personal lines will be rendered differently.
*/

char* httpc_duplines(client_t* client) {
  int size=1024;
  char* result = (char*)malloc(size);
  char* hash[client->session->box->maxlines];
  int len = httpc_db_dup_lines(client, hash, client->session->box->maxlines);
  *result = 0;
  while (--len >= 0) {
    conf_t* context = c_parse(hash[len]);
    char* line = c_first_child_value_named(context, -1, "line");
    if (context && line) {
      int line_is_visible = 0;
      int user_is_recipient = 0;
      char* recipient, * user, * cookie;
      recipient = c_first_child_value_named(context, -1, "recipient");
      user      = c_first_child_value_named(context, -1, "user");
      cookie    = c_first_child_value_named(context, -1, "cookie");
      if (recipient && client->session->user) {
        if (!strcmp(recipient, client->session->user)) {
          line_is_visible = 1;
          user_is_recipient = 1;
        }
        if (user && !strcmp(user, client->session->user)) {
          line_is_visible = 1;
        }
      } else {
        line_is_visible = 1;
      }
      if (line_is_visible) {
        char* linebuf;
        if (user_is_recipient) {
          linebuf = httpc_format_dup(client, client->session->box->message,
                                     context,
                                     client->session->var,
                                     client->session->box->var);
        } else if (httpcP_is_personal(client, cookie, user)) {
          linebuf = httpc_format_dup(client, client->session->box->personal,
                                     context,
                                     client->session->var,
                                     client->session->box->var);
        } else {
          linebuf = httpc_format_dup(client, client->session->box->line,
                                     context,
                                     client->session->var,
                                     client->session->box->var);
        }
        result = dupcat(result, &size, linebuf);
        free(linebuf);
      }
      c_close(context);
    }
    free(hash[len]);
  }
  result = dupcat(result, &size, "<a name=focus></a>");
  return result;
}

/*
* Returns a malloced string containing all users in this box.
*/

/*
char* httpc_dup_users(httpc_box_t* box) {
  int size = 1024;
  httpc_t* t = box->server->httpc;
  char* result = (char*)malloc(size);
  strcpy(result, "Users: ");
  if (pthread_mutex_lock(&(t->sessions.mutex)) == 0) {
    int i;
    for (i=0; i<t->sessions.count; i++) {
      if (t->sessions.list[i]->box == box) {
        char* user = t->sessions.list[i]->user;
        if (user) {
          result = dupcat(result, &size, user);
          result = dupcat(result, &size, ", ");
        }
      }
    }
    pthread_mutex_unlock(&(t->sessions.mutex));
  }
  return result;
}
*/
