/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_client.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <stdarg.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include <netinet/tcp.h>

#include "httpc.h"
#include "httpc_inet.h"
#include "httpc_util.h"
#include "httpc_box.h"
#include "httpc_format.h"

static const char serverheader[] = "Server: httpcd 0.2\r\n";

static char ieflush[] = 
  "<!--"
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "===================================================================="
  "-->";

int httpc_print(client_t* c, char* page) {
  if (fdwrite(c->fd, page)) {
#ifdef _DEBUG
    perror("httpc_print:fdwrite");
#endif
    return -1;
  }
  return 0;
}

int httpc_flush(client_t* c) {
  int i, j;
  i = 0; j = sizeof(i);
  if (getsockopt(c->fd, SOL_SOCKET, SO_SNDBUF, (char *) &i,
                 (socklen_t *) &j) < 0) {
    perror("getsockopt SO_SNDBUF");
  }
  if (j != sizeof(i)) {
    fprintf(stderr, "getsockopt size=%d!\n", j);
  }
  if (fdwrite(c->fd, ieflush)) { perror("httpc_flush:fdwrite"); return -1; }
/*
  j = 2;
  if (setsockopt(c->fd, SOL_SOCKET, SO_SNDBUF, (char *) &j,
			sizeof(int)) < 0) {
    perror("setsockopt SO_SNDBUF");
  }
  if (fdwrite(c->fd, "\r\n")) { perror("httpc_flush:fdwrite"); return -1; }
  if (setsockopt(c->fd, SOL_SOCKET, SO_SNDBUF, (char *) &i,
                 sizeof(int)) < 0) {
    perror("setsockopt SO_SNDBUF");   
  }
*/
  return 0;
}

static int counter = 0;

int httpc_print_none(client_t* client) {
  return (fdprintf(client->fd,
                   "HTTP/1.0 200 OK\r\n"
                   "Content-type: text/html\r\n"
                   "%s"
                   "\r\n"
                   "<html><!-- %d --></html>"
                   , serverheader
                   , counter++));
}

int httpc_print_unauthorized(client_t* client) {
  return (fdprintf(client->fd,
                   "HTTP/1.0 401 Unauthorized\r\n"
                   "Content-type: text/html\r\n"
                   "%s"
                   "WWW-Authenticate: Basic realm=\"Chatserver %s\"\r\n"
                   "\r\n"
                   "<html><!-- %d --><body>Server '%s' requires authentication."
                   "</body></html>\n"
                   , serverheader
                   , client->server->name
                   , counter++
                   , client->server->name
                   ));
}

int httpc_print_error(client_t* c, char* message) {
  if (fdprintf(c->fd,
               "HTTP/1.0 200 OK\r\n"
               "Content-type: text/html\r\n"
               "%s"
               "\r\n"
               "<html><!-- %d --><body>Error: %s</body></html>\n"
               , serverheader
               , counter++
               , message
               )) return -1;
  return 0;
}

int httpcP_print_head(client_t* c) {
  if (c->session->box->head) {
    if (httpc_format_stream(c, c->session->box->head,
        c->session->var, c->session->box->var, c->request.query)) {
      return -1;
    }
  }
  return 0;
}

int httpcP_print_bodyprefix(client_t* c) {
  if (c->session->box->bodyprefix) {
    if (httpc_format_stream(c, c->session->box->bodyprefix,
        c->session->var, c->session->box->var, c->request.query)) {
      return -1;
    }
  }
  return 0;
}

int httpcP_print_bodypostfix(client_t* c) {
  if (c->session->box->bodypostfix) {
    if (httpc_format_stream(c, c->session->box->bodypostfix,
        c->session->var, c->session->box->var, c->request.query)) {
      return -1;
    }
  }
  return 0;
}

int httpc_print_chatlines_push(client_t* c, char* page) {
  if (fdprintf(c->fd,
               "\r\n--==========oi==========\r\n"
               "HTTP/1.0 200 OK\r\n"
               "Content-type: text/html\r\n"
               "%s"
               "\r\n<html><!-- %d --><head>"
               , serverheader
               , counter++
               )) return -1;
  if (httpcP_print_head(c)) return -1;
  if (httpc_print(c, "</head><body>\n")) return -1;
  if (httpcP_print_bodyprefix(c)) return -1;
  if (httpc_print(c, page)) return -1;
  if (httpcP_print_bodypostfix(c)) return -1;
  if (httpc_print(c, 
                  "<script language=javascript><!--\n"
                  "dh = document.height;\n"
                  "wh = window.innerHeight;\n"
                  "var sh = dh - wh;\n"
                  "if (sh > 0) window.setTimeout("
                  "'window.scrollTo(0, sh)', 200);\n"
                  "//-->\n</script>\n")) return -1;
  if (httpc_print(c, "</html>")) return -1;
  return 0;
}

int httpc_print_chatlines_js(client_t* c, char* page) {
  if (fdprintf(c->fd,
               "HTTP/1.0 200 OK\r\n"
               "Content-type: text/html\r\n"
               "%s"
               "\r\n"
               "<html><!-- %d --><head>"
               , serverheader
               , counter++
               )) return -1;
  if (httpcP_print_head(c)) return -1;
#ifdef MSIE_ONLOAD
  if (httpc_print(c, 
      "</head><body onLoad=\"window.location.reload();\">\n")) return -1;
#else
  if (httpc_print(c, "</head><body>\n")) return -1;
#endif
  if (httpcP_print_bodyprefix(c)) return -1;
  if (httpc_print(c, page)) return -1;
  if (httpcP_print_bodypostfix(c)) return -1;
  return 0;
}

int httpc_print_chatlines_rest(client_t* c, char* page) {
  char buffer[128];
  time_t t = time(NULL);
  struct tm* tm = gmtime(&t);
  strftime(buffer, 128, "%a, %d %b %Y %H:%M:%S GMT", tm);
  if (fdprintf(c->fd,
               "HTTP/1.0 200 OK\r\n"
               "Content-type: text/html\r\n"
               "%s"
               "Date: %s\r\n"
               "Last-modified: %s\r\n"
               "Expires: %s\r\n"
               "\r\n"
               "<html><!-- %d --><head>"
               , serverheader
               , buffer
               , buffer
               , buffer
               , counter++
               )) return -1;
  if (httpcP_print_head(c)) return -1;
  if (fdprintf(c->fd, "<meta HTTP-EQUIV=\"REFRESH\""
                      " content=\"7; URL=/%s\">\n",
                      c->session->cookie)) return -1;
  if (httpc_print(c, "</head><body>\n")) return -1;
  if (httpcP_print_bodyprefix(c)) return -1;
  if (httpc_print(c, page)) return -1;
  if (httpcP_print_bodypostfix(c)) return -1;
  if (httpc_print(c, "</body></html>\n")) return -1;
  return 0;
}

int httpc_chatbox2newclient(client_t* c, char* page, int cclose) {
  switch (c->session->browser) {
  case CAPABLE:
    if (fdprintf(c->fd,
                 "HTTP/1.0 200 OK\r\n"
                 "Content-type: multipart/x-mixed-replace;"
                 " boundary=\"==========oi==========\"\r\n"
                 "\r\n")) return -1;
    if (httpc_print_chatlines_push(c, page)) return -1;
    c->sucking = 1;
    return (0 | cclose);
  case JAVASCRIPT_CAPABLE:
    if (httpc_print_chatlines_js(c, page)) return -1;
    c->sucking = 1;
    return (httpc_flush(c) | 0 | cclose);
  case INCAPABLE:
    if (httpc_print_chatlines_rest(c, page)) return -1;
  }
  return 1;
}

int httpc_chatbox2allbut(client_t* c) {
  if (pthread_mutex_lock(&(c->server->clients.mutex)) == 0) {
    int j;
    for (j=0; j<c->server->clients.count; j++) {
      client_t* client = c->server->clients.list[j];
      if (client == c) continue;
      if (client->sucking &&
          !strcmp(client->session->boxname, c->session->boxname) &&
          client->writestate < SELECTABLE) {
        client->writestate = SELECTABLE;
      }
    }
    pthread_mutex_unlock(&(c->server->clients.mutex));
    return 0;
  }
  return -1;
}
