/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_handler.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <unistd.h>
#include <fcntl.h>
#include <time.h>

#include "httpc_handler.h"
#include "httpc_client.h"
#include "httpc_box.h"
#include "httpc_util.h"
#include "httpc_session.h"
#include "httpc_format.h"
#include "base64.h"

static inline char * skiphex(const char * path, char* buffer) {
  char * e = (char *) path;
  if (*e++ == '/') {
    int n = strspn(e, "0123456789ABCDEFabcdef");
    if (e[n]=='/' && n<33) {
      strncpy(buffer, e, n);
      buffer[n] = 0;
      return e+n;
    }
  }
  return (char *) path;
}

int httpcP_print_result_file(client_t* c, char* file) {
  int n;
  if ((n = c_first_child_named(c->server->files, -1, file)) != -1) {
    httpc_format_stream(c, c_value(c->server->files, n),
                        c->session->var, NULL, NULL);
    return 0;
  }
  return -1;
}

int httpcP_is_auth_table(client_t* c) {
  a_ngrid_t* table = c->server->auth.types.pwdf.table;
  if (table) {
    int n = c_first_child_casenamed(c->request.headers, -1, "Authorization");
    if (n != -1) {
      char* value = c_value(c->request.headers, n);
      if (value && !strncmp(value, "Basic ", 6)) {
        char* cookie = value + 6;
        int l = b64_decodedlength(cookie);
        char buf[l+1];
        if (buf) {
          char* colon;
          b64_decode(cookie, buf);
          buf[l] = 0;
          colon = strchr(buf, ':');
          if (colon) {
            int row;
            *colon = 0;
            row = g_findincol(table, 0, buf);
            if (row != -1) {
              if (!strcmp(g_getcell(table, row, 1), colon+1)) {
                httpc_set_user(c->session, buf);
                return 1;
              }
            }
          }
        }
      }
    }
  }
  return 0;
}

int httpcP_is_authorized(client_t* c) {
  if (!c->server->auth.use_auth) {
    return 1;
  } else {
    int result = 0;
    switch (c->server->auth.type) {
    case HTTPC_AUTH_TABLE:
      result = httpcP_is_auth_table(c);
      break;
    }
    if (!result) httpc_print_unauthorized(c);
    return result;
  }
}

browser_t httpcP_find_browser(client_t* c) {
  int n = c_first_child_casenamed(c->request.headers, -1, "User-Agent");
  if (n != -1) {
    char* useragent = c_value(c->request.headers, n);
    if (!strncasecmp(useragent, "Mozilla", 7)) {
      if (strstr(useragent, "MSIE")) {
        return JAVASCRIPT_CAPABLE;
      } else {
        return CAPABLE;
      }
    }
  }
  return INCAPABLE;
}

int httpcP_assign_session(client_t* c, char* cookie, char* path) {
  if (!(c->session = httpc_lookup_session(c->server->httpc, cookie))) {
    if (!(c->session = httpc_add_session(c->server->httpc))) {
      httpc_print_error(c, "No session could be created.");
      return -1;
    }
//    memset(c->session->boxname, 0, BOXNAMESIZE);
    if (c->server->musthavecookie) {
      httpc_print_error(c, "No session could be assigned.");
      return -1;
    }
    c->session->browser = httpcP_find_browser(c);
    c->session->box = &(c->server->boxes.list[0]);
    httpc_set_boxname(c->session, "");
/*
    if ((users = httpc_dup_users(c->session->box))) {
      httpc_add_line(c, users, NULL);
      free(users);
    }
*/
    if (cookie && c->server->useforeigncookie) {
      httpc_set_cookie(c->session, cookie);
    } else {
      httpc_create_cookie(c->session);
    }
    memcpy(&(c->session->address),
           &(c->address),
           sizeof(struct in_addr));
  }
  return 0;
}

int httpcP_client_session_exchange(client_t* c) {
  if (memcmp(&(c->address),
             &(c->session->address),
             sizeof(struct in_addr))) {
    return -1;
  }
  return 0;
}

int httpcP_handle_http_request(client_t* c) {
  int cclose        = 0;
  int hasline       = 0;
  char* page        = NULL;
  int resultprinted = 0;
  int result        = 0;
  char* line;
  char* path = skip_whitespace(c->request.path);
  char* p = path;
  char buffer[34], * cookie = buffer;
  if ((path = skiphex(path, buffer)) == p) {
    cookie = NULL;
  }
  if (httpcP_assign_session(c, cookie, path)) {
    return -1;
  }
  if (httpcP_client_session_exchange(c)) {
    httpc_print_error(c, "Session overruled client.");
    return -1;
  }
  //-- increases client reference count of the session
  httpc_session_inc(c->server->httpc, c->session);
  if (!httpcP_is_authorized(c)) return -1;
  if ((line =
       c_first_child_value_named(c->request.query, -1, "line")) != NULL &&
       strlen(line)) {
    hasline = 1;
    httpc_interpret_line(c, line);
  }
  if (strcmp(path, "/")) {
    if (httpcP_print_result_file(c, path+1)) {
      httpc_print_error(c, "Unknown path value.");
    }
    result=1;
    resultprinted = 1;
  }
  if (!resultprinted) {
    if (!page) page = httpc_duplines(c);
    if (page) {
      result = httpc_chatbox2newclient(c, page, cclose);
    }
    resultprinted = 1;
  }
  if (hasline) {
    httpc_chatbox2allbut(c);
  }
  if (page) free(page);
  return result;
}

char* httpcP_dupheaders(int fd) {
  int i = 0, size = 128;
  char c, * result = (char*)malloc(size);
  while (read(fd, &c, 1) > 0) {
    if (size > MAXHEADERSIZE) {
      free(result);
      return NULL;
    }
    result[i] = c;
    if (++i >= size) {
      result = (char*)realloc(result, (size += 1024));
      if (!result) return NULL;
    }
    if (c == '\n') {
      if (i>0 && result[i-2] == '\n') {
        result[i] = 0;
        return result;
      }
      if (i>2 && result[i-2] == '\r' &&
                 result[i-3] == '\n' &&
                 result[i-4] == '\r') {
        result[i] = 0;
        return result;
      }
    } else if (c == '\r') {
      if (i>0 && result[i-2] == '\r') {
        result[i] = 0;
        return result;
      }
    }
  }
  free(result);
  return NULL;
}

/*
* Entry point for a http request (read/write) handler.
*/

#include "httpc_parse.h"

void httpc_handle_http_request(client_t* c) {
  char* headers = httpcP_dupheaders(c->fd);
  if (headers) {

#ifdef _DEBUG
    httpc_parse_request(c->address, c->port, &(c->request), headers, c->fd);
#else
    httpc_parse_request(c->address, &(c->request), headers, c->fd);
#endif
    free(headers);

    if (httpcP_handle_http_request(c)) {
      httpc_close_client(c);
    } else {
      c->readstate = SELECTABLE;
    }
    httpc_free_request(&(c->request));
    return;
  }
  httpc_close_client(c);
}

/*
* Writes a box to a sucking client; entry point for a write handler.
*/

#ifndef MSIE_ONLOAD
static int msiespoof = 0;
#endif

void httpc_handle_write(client_t* c) {
  char* page;
  switch (c->session->browser) {
  case CAPABLE:
    if ((page = httpc_duplines(c))) {
      if (httpc_print_chatlines_push(c, page)) {
        httpc_close_client(c);
      }
      free(page);
    }
    break;
  case JAVASCRIPT_CAPABLE:
#ifndef MSIE_ONLOAD
    fdprintf(c->fd,
             "<script language=javascript>\n"
             "<!--\n"
             "location.href='/%s/?junk=%u%d#focus';\n"
             "//-->\n"
             "</script></html>\n"
             , c->session->cookie
             , time(NULL)
             , msiespoof++
            );
#endif
    fdprintf(c->fd,
             "</body></html>\n");
    httpc_close_client(c);
    break;
  case INCAPABLE:
    // impossible
    fprintf(stderr, "!! Trying to write chatbox to incapable browser.\n");
    httpc_close_client(c);
    break;
  }
  c->writestate = SILENT;
}
