/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_parse.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "httpc_parse.h"
#include "httpc_util.h"

//#define _DEBUG1

#ifdef _DEBUG1
void httpcP_print_request(struct httprequest* req) {
  char* m[] = { "", "GET", "POST" };
  int n;
  fprintf(stderr, "\nmethod '%s' path '%s' %d.%d\n",
                  m[req->method], req->path,
                  req->proto_major, req->proto_minor);
  fprintf(stderr, "----\n");
  n = c_first_child(req->query, -1);
  while (n != -1) {
    fprintf(stderr, "'%s' = '%s'\n",
                    c_name(req->query, n), c_value(req->query, n));
    n = c_next_brother(req->query, n);
  }
  fprintf(stderr, "----\n");
  n = c_first_child(req->headers, -1);
  while (n != -1) {
    fprintf(stderr, "'%s' = '%s'\n",
                    c_name(req->headers, n), c_value(req->headers, n));
    n = c_next_brother(req->headers, n);
  }
  fprintf(stderr, "----\n");
}
#endif


char* skip_line(char* start, char** next) {
  char* s = start;
  while (*s && !strchr("\n\r", *s)) {
    ++s;
  }
  if (!*s) {
    *next = NULL;
    return s;
  }
  if (s[0] == '\r' && s[1] == '\n') {
    *next = s+2;
  } else {
    *next = s+1;
  }
  while (s >= start && strchr(" \n\r\t", *s)) --s;
  if (s == start) return NULL;
  return s+1;
}

char* html_allowed[] = {
  "b",
  "i",
  "u",
  NULL
};

void httpcP_peel_query(struct httprequest* req, char* query) {
  char* n = query, * v = NULL, * s = query;
  for (;;) {
    if (*s == '\0' || *s == '&') {
      int breakme = ((*s == '\0') ? 1 : 0);
      *s = 0;
      if (v) {
        char* name = httpc_unescape(n);
        char* value = httpc_unescape(v);
        cgi_tagfilter(value, html_allowed);
        c_add_child(req->query, -1, name, value);
        v[-1] = '=';
      } else {
        char* value = httpc_unescape(n);
        cgi_tagfilter(value, html_allowed);
        c_add_child(req->query, -1, "line", value);
      }
      if (breakme) break;
      *s = '&';
      n = s+1;
      v = NULL;
    } else if (*s == '=') {
      *s = 0;
      v = s+1;
    }
    ++s;
  }
}

void trim_protocol(struct httprequest* req, char* line) {
  char* proto = strstr(line, "HTTP/");
  if (proto) {
    size_t s = strspn(proto+5, "0123456789.");
    if (s && s == strlen(proto+5)) {
      sscanf(proto+5, "%hd.%hd", &(req->proto_major), &(req->proto_minor));
      *proto = 0;
      trim(line);
    }
  }
}

void trim_query(struct httprequest* req, char* line) {
  char* query = strchr(line, '?');
  if (query) {
    *query = 0;
    if (!req->query) req->query = c_create();
    httpcP_peel_query(req, ++query);
  }
}

int httpcP_parse_first_line(struct httprequest* req, char* line) {
  if (!strncasecmp(line, "GET ", 4)) {
    req->method = HTTP_METHOD_GET;
    line += 4;
  } else if (!strncasecmp(line, "POST ", 5)) {
    req->method = HTTP_METHOD_POST;
    line += 5;
  }
  line = skip_whitespace(line);
  trim_protocol(req, line);
  trim_query(req, line);
  if (strlen(line) >= HTTPC_MAXPATH) {
    return -1;
  }
  strcpy(req->path, line);
  return 0;
}

int httpcP_add_header(struct httprequest* req, char* line) {
  char* colon = strchr(line, ':');
  if (colon) {
    char* value = skip_whitespace(colon+1);
    if (value) {
      *colon = 0;
      c_add_child(req->headers, -1, line, value);
      *colon = ':';
    }
  }
  return 0;
}

#define HTTP_POST_TYPE_x_www_form_urlencoded  1
#define HTTP_POST_TYPE_multipart_form_data    2

int httpcP_handle_POST(struct httprequest* req, int fd) {
  char* buffer;
  int type = HTTP_POST_TYPE_x_www_form_urlencoded;
  int n = c_first_child_named(req->headers, -1, "Content-Type");
  if (n != -1) {
    char* value = c_value(req->headers, n);
    if (value && !strcmp(value, "multipart/form-data")) {
      type = HTTP_POST_TYPE_multipart_form_data;
      //.. should be supported really
      return -1;
    }
  }
  n = c_first_child_casenamed(req->headers, -1, "Content-Length");
  if (n == -1) {
    // some browsers do not send content length with a POST,
    // contrary to what RFC 1945 says.
    if (!(buffer = fdstrdup(fd, MAXPOSTSIZE))) return -1;
  } else {
    int rl, l = c_value_int(req->headers, n);
    if (l <= 0 || l > MAXPOSTSIZE) return -1;
    if (!(buffer = (char*)malloc(l+1))) return -1;
    if ((rl = slurp(fd, buffer, l)) == l) {
      buffer[l] = 0;
    } else {
      fprintf(stderr, "Content length %d != actual bytes read (%d).\n"
                      , l, rl);
      free(buffer);
      return -1;
    }
  }
  switch (type) {
  case HTTP_POST_TYPE_x_www_form_urlencoded:
    if (!req->query) req->query = c_create();
    httpcP_peel_query(req, buffer);
    break;
  case HTTP_POST_TYPE_multipart_form_data:
    //..
    break;
  }
  return 0;
}

#ifdef _DEBUG
int httpc_parse_request(struct in_addr ina, short port,
                        struct httprequest* req, char* headers, int fd) {
#else
int httpc_parse_request(struct in_addr ina,
                        struct httprequest* req, char* headers, int fd) {
#endif
  char* eol, * begin, * next;
  memset(req, 0, sizeof(*req));
  if ((eol = skip_line(headers, &next))) {
    char swap = *eol;
    *eol = 0;
    fprintf(stderr, "-%s %s\n", inet_ntoa(ina), headers);
    httpcP_parse_first_line(req, headers);
    *eol = swap;
    if (next) {
      req->headers = c_create();
      begin = next;
      while ((eol = skip_line(begin, &next))) {
        swap = *eol;
        *eol = 0;
        httpcP_add_header(req, begin);
        *eol = swap;
        if (!next) break;
        begin = next;
      }
    }
  }
  if (req->method == HTTP_METHOD_POST) {
    if (httpcP_handle_POST(req, fd)) return -1;
  }
#ifdef _DEBUG1
  httpcP_print_request(req);
#endif
  return 0;
}

void httpc_free_request(struct httprequest* req) {
  if (req->headers) c_close(req->headers);
  if (req->query) c_close(req->query);
}
