/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_session.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include "httpc_session.h"

/*
* In general, during normal execution time, session are added by
* handler-threads, and removed by watcher-threads (if their client
* pointer is NULL and their timestamp ran out).
*/

session_t* httpc_add_session(httpc_t* t) {
  session_t* session = (session_t*)malloc(sizeof(session_t));
  if (session) {
    memset(session, 0, sizeof(*session));
    if (pthread_mutex_lock(&(t->sessions.mutex)) == 0) {
      t->sessions.list =
          (session_t**)realloc(t->sessions.list,
          sizeof(session_t*) * (t->sessions.count+1));
      if (!t->sessions.list) {
        free(session);
        return NULL;
      }
      t->sessions.list[(t->sessions.count)++] = session;
      pthread_mutex_unlock(&(t->sessions.mutex));
      session->var = c_create();
      session->timestamp = time(NULL);
    } else {
      free(session);
      return NULL;
    }
  }
  return session;
}

void httpc_remove_session_safe(httpc_t* t, session_t* s) {
  int i;
  for (i=0; i<t->sessions.count; i++) {
    if (t->sessions.list[i] == s) {
      if (i < t->sessions.count-1) {
        memmove(&(t->sessions.list[i]),
                &(t->sessions.list[i+1]),
                sizeof(session_t*) * (t->sessions.count - (i+1)));
      }
      t->sessions.count--;
      if (s->var) c_close(s->var);
      free(s);
      break;
    }
  }
}

void httpc_remove_session(httpc_t* t, session_t* s) {
  if (pthread_mutex_lock(&(t->sessions.mutex)) == 0) {
    httpc_remove_session_safe(t, s);
    pthread_mutex_unlock(&(t->sessions.mutex));
  }
}

session_t* httpc_lookup_session(httpc_t* t, char* cookie) {
  session_t* result = NULL;
  if (!cookie) return NULL;
  if (pthread_mutex_lock(&(t->sessions.mutex)) == 0) {
    int i;
    for (i=0; i<t->sessions.count; i++) {
      if (!strcmp(t->sessions.list[i]->cookie, cookie)) {
        result = t->sessions.list[i];
        break;
      }
    }
    pthread_mutex_unlock(&(t->sessions.mutex));
  }
  return result;
}

static int counter = 0;

int httpc_create_cookie(session_t* s) {
  int i;
  time_t t = time(NULL);
  char cookie[33], hexstring[] = "0123456789abcdef";
  unsigned char cookiebytes[16];
  memset(cookiebytes, 0, 16);
  memcpy(cookiebytes, &t, sizeof(time_t));
  counter++;
  memcpy(cookiebytes + sizeof(time_t), &counter, sizeof(int));
  for (i=0; i<16; i++) {
    cookie[i*2] = hexstring[ (cookiebytes[i] / 0x10) ];
    cookie[(i*2)+1] = hexstring[ (cookiebytes[i] % 0x10) ];
  }
  cookie[32] = 0;
  return httpc_set_cookie(s, cookie);
}

int httpc_set_cookie(session_t* s, char* cookie) {
  int n = c_first_child_named(s->var, -1, "cookie");
  if (n == -1) {
    n = c_add_child(s->var, -1, "cookie", cookie);
  } else {
    c_set(s->var, n, cookie);
  }
  s->cookie = c_value(s->var, n);
  return 0;
}

int httpc_set_user(session_t* s, char* user) {
  int n = c_first_child_named(s->var, -1, "user");
  if (n == -1) {
    n = c_add_child(s->var, -1, "user", user);
  } else {
    c_set(s->var, n, user);
  }
  s->user = c_value(s->var, n);
  return 0;
}

int httpc_set_boxname(session_t* s, char* boxname) {
  int n = c_first_child_named(s->var, -1, "boxname");
  if (n == -1) {
    n = c_add_child(s->var, -1, "boxname", boxname);
  } else {
    c_set(s->var, n, boxname);
  }
  s->boxname = c_value(s->var, n);
  return 0;
}

void httpc_session_inc(httpc_t* t, session_t* s) {
  if (pthread_mutex_lock(&(t->sessions.mutex)) == 0) {
    s->timestamp = time(NULL);
    s->clientrefcount++;
    pthread_mutex_unlock(&(t->sessions.mutex));
  }
}

void httpc_session_dec(httpc_t* t, session_t* s) {
  if (pthread_mutex_lock(&(t->sessions.mutex)) == 0) {
    s->timestamp = time(NULL);
    s->clientrefcount--;
    pthread_mutex_unlock(&(t->sessions.mutex));
  }
}
