/*
 * (C) 2000 ORG.NET bv, The Hague, Netherlands.
 * credits to "Kees Jan Hermans" <kees@orgnet.nl>
 * many new bugs added by "Niels Poppe" <niels@netbox.org>
 * $Id: httpc_util.c,v 1.2 2000/11/06 11:06:59 niels Exp $
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "httpc_util.h"

int fdprintf(int fd, char* fmt, ...) {
  int len, n, off=0;
  char line[4096];
  va_list ap;
  va_start(ap, fmt);
  vsnprintf(line, 4096, fmt, ap);
  va_end(ap);
  if ((len = strlen(line)) == 0) return 0;
  while ((n = write(fd, line+off, len-off)) > 0) {
    if ((off += n) == len) break;
  }
  return ((n < 0) ? n : 0);
}

int fdwrite(int fd, char* str) {
  int len = strlen(str);
  int n, off = 0;
  while ((n = write(fd, str+off, len-off)) > 0) {
    if ((off += n) == len) break;
  }
  return ((n < 0) ? n : 0);
}

int httpcP_hex2num(char hex) {
  if (hex >= '0' && hex <= '9') {
    return hex - '0';
  }
  if (hex >= 'a' && hex <= 'f') {
    return hex - 'a' + 10;
  }
  if (hex >= 'A' && hex <= 'F') {
    return hex - 'A' + 10;
  }
  return 0;
} 
  
char* httpc_unescape(char* line) {
  char* org = line, *dst = line;
  while (*org) {
    if (*org == '+') {
      *dst++ = ' ';
    } else if (*org == '%') {
      char max = *(++org);
      char min = *(++org);
      *dst++ = (char)(httpcP_hex2num(max) * 0x10 + httpcP_hex2num(min));
    } else {
      *dst++ = *org;
    }
    ++org;
  }
  *dst = 0;
  return line;
} 

char* fdstrdupP(int fd, int maxsize) {
  int n, off = 0;
  char* result = NULL;
  for (;;) {
    if ((result = (char*)realloc(result, off + 1025)) == NULL) {
      break;
    }
    if ((n = read(fd, result+off, 1024)) < 0) {
      free(result);
      result = NULL;
      break;
    }
    if ((off += n) > maxsize) {
       free(result);
      result = NULL;
      break;
    }
    if (n == 0) {
      result[off] = 0;
      break;
    }
  }
  return result;
}

char* fstrdup(char* path, int maxsize) {
  int fd = open(path, O_RDONLY);
  if (fd >= 0) {
    char* result = fdstrdupP(fd, maxsize);
    close(fd);
    return result;
  }
  return NULL;
}

char* fdstrdup(int fd, int maxsize) {
  char* result;
  int flags = fcntl(fd, F_GETFL);
  fcntl(fd, F_SETFL, O_NONBLOCK);
  result = fdstrdupP(fd, maxsize);
  fcntl(fd, F_SETFL, flags);
  return result;
}

int slurp(int fd, char* buffer, int size) {
  int off = 0;
  for (;;) {
    int l, len = 1024;
    if (off + len > size) len = size - off;
    if (len == 0) return size;
    if ((l = read(fd, buffer+off, len)) < 0) return l;
    off += l;
  }
}

char* dupcat(char* buffer, int* size, char* orig) {
  if (buffer && *size >= 0) {
    if (orig) {
      int len0 = strlen(buffer);
      int len1 = strlen(orig);
      while (len0 + len1 + 1 > *size) {
        buffer = (char*)realloc(buffer, (*size += 1024));
        if (!buffer) return buffer;
      }
      strcpy(buffer + len0, orig);
    }
    return buffer;
  }
  return NULL;
}

char* trim(char* string) {
  int l = strlen(string);
  while (l >= 0 && string[l] <= ' ') {
    string[l--] = 0;
  }
  return string;
}

char* skip_whitespace(char* string) {
  int i = 0, l = strlen(string);
  while (i < l && strchr(" \n\r\t", string[i])) {
    ++i;
  }
  if (i == l) return NULL;
  return string + i;
}

char* skip_identifier(char* string) {
  char* s = string;
  if (*s == '\'' ||
      *s == '\"') {
    char c = *s++;
    while (*s && *s != c) ++s;
    if (*s == c) return s+1;
  }
  s = string;
  while (*s && !strchr(" \n\r\t", *s)) ++s;
  return s;
}

// stolen from cgilib.c and strskip.c

char* skip_ontok(char* string, char* delim) {
  while (*string && !strchr(delim, *string)) ++string;
  return string;
}

int cgi_tagfilter(char* s, char** html_allowed) {
  while (*s) {
    if (*s == '<') {
      char* start = s;
      char* end   = strchr(start, '>');
      if (end) {
        char* tagstart, * tagend, replace, ** a;
        int pad = 1;
        tagstart = skip_whitespace(start+1);
        if (*tagstart == '/') ++tagstart;
        tagend  = skip_ontok(tagstart, " \n\r\t>");
        replace = *tagend;
        *tagend = '\0';
        a = html_allowed;
        while (*a) {
          if (!strcasecmp(*a, tagstart)) {
            pad = 0;
          }
          ++a;
        }
        *tagend = replace;
        if (pad) {
          int i; for (i=0; i< (1+end-start); i++) {
            start[i] = ' ';
          }
        }
        s = end;
      } else return -1;
    }
    ++s;
  }
  return 0;
}

